"""Thermostat device types."""

from datetime import datetime
import logging

from ..aldb.aldb import ALDB
from ..aldb.aldb_battery import ALDBBattery
from ..config import (
    BACKLIGHT,
    BUTTON_LOCK_ON,
    CELSIUS,
    CHANGE_DELAY,
    HUMIDITY_OFFSET,
    KEY_BEEP_ON,
    LED_ON,
    OPERATING_FLAGS,
    PROGRAM_LOCK_ON,
    TEMP_OFFSET,
    TIME_24_HOUR_FORMAT,
)
from ..config.op_flag_property import OpFlagProperty
from ..constants import PropertyType, ResponseStatus, ThermostatMode
from ..default_link import DefaultLink
from ..groups import (
    COOL_SET_POINT,
    COOLING,
    DEHUMIDIFYING,
    FAN_MODE,
    HEAT_SET_POINT,
    HEATING,
    HUMIDIFYING,
    HUMIDITY,
    HUMIDITY_HIGH,
    HUMIDITY_LOW,
    SYSTEM_MODE,
    TEMPERATURE,
)
from ..groups.on_off import OnOff
from ..groups.thermostat import FanMode, Humidity, SetPoint, SystemMode, Temperature
from ..handlers.from_device.thermostat_cool_set_point import (
    ThermostatCoolSetPointHandler,
)
from ..handlers.from_device.thermostat_heat_set_point import (
    ThermostatHeatSetPointHandler,
)
from ..handlers.from_device.thermostat_humidity import ThermostatHumidityHandler
from ..handlers.from_device.thermostat_mode import ThermostatModeHandler
from ..handlers.from_device.thermostat_temperature import ThermostatTemperatureHandler
from ..handlers.to_device.extended_set import ExtendedSetCommand
from ..handlers.to_device.thermostat_cool_set_point import ThermostatCoolSetPointCommand
from ..handlers.to_device.thermostat_heat_set_point import ThermostatHeatSetPointCommand
from ..handlers.to_device.thermostat_mode import ThermostatModeCommand
from ..managers.link_manager.default_links import async_add_default_links
from ..managers.on_level_manager import OnLevelManager
from ..managers.thermostat_status_manager import GetThermostatStatus
from ..utils import multiple_status, to_fahrenheit
from .battery_base import BatteryDeviceBase
from .device_base import Device
from .device_commands import STATUS_COMMAND

_LOGGER = logging.getLogger(__name__)

GRP_COOL_ON = 1
GRP_HEAT_ON = 2
GRP_HUMID_HI_ON = 3
GRP_HUMID_LO_ON = 4
GRP_NOTIFY = 0xEF

GRP_TEMP = 10
GRP_HUMID = 11
GRP_SYS_MODE = 12
GRP_FAN_MODE = 13
GRP_COOL_SP = 14
GRP_HEAT_SP = 15
GRP_HUMID_HI_SP = 16
GRP_HUMID_LO_SP = 17

OP_FLAG_POS = 13


class ClimateControl_Thermostat(Device):
    """Thermostat device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the Thermostat class."""
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
        )
        self._aldb = ALDB(self._address, mem_addr=0x1FFF)

    async def async_set_cool_set_point(self, temperature):
        """Set the cool set point."""
        temperature = max(1, min(temperature, 127))
        return await self._handlers["cool_set_point_command"].async_send(temperature)

    async def async_set_heat_set_point(self, temperature):
        """Set the cool set point."""
        temperature = max(1, min(temperature, 127))
        return await self._handlers["heat_set_point_command"].async_send(temperature)

    async def async_set_humidity_high_set_point(self, humidity):
        """Set the humidity high set point."""
        humidity = min(humidity, 99)
        cmd_status = await self._handlers["humidity_high_command"].async_send(
            data3=humidity
        )
        if cmd_status == ResponseStatus.SUCCESS:
            self._groups[GRP_HUMID_HI_SP].value = humidity
        return cmd_status

    async def async_set_humidity_low_set_point(self, humidity):
        """Set the humidity low set point."""
        humidity = max(humidity, 1)
        cmd_status = await self._handlers["humidity_low_command"].async_send(
            data3=humidity
        )
        if cmd_status == ResponseStatus.SUCCESS:
            self._groups[GRP_HUMID_LO_SP].value = humidity
        return cmd_status

    async def async_set_mode(self, thermostat_mode):
        """Set the thermostat mode."""
        return await self._handlers["mode_command"].async_send(thermostat_mode)

    async def async_set_master(self, master):
        """Set the thermostat master mode."""
        return await self._handlers["set_master"].async_send(data3=master)

    async def async_set_notify_changes(self):
        """Set the thermostat to notify of changes."""
        return await self._handlers["notify_changes_command"].async_send()

    async def async_set_day_time(self):
        """Set the thermostat day of the week and time."""
        curr_time = datetime.now()
        day = 0 if curr_time.weekday() == 6 else curr_time.weekday() + 1
        set_time_command = ExtendedSetCommand(
            self._address, cmd2=0x02, data1=0x02, data2=day
        )
        return await set_time_command.async_send(
            data3=curr_time.hour, data4=curr_time.minute, data5=curr_time.second
        )

    async def async_add_default_links(self):
        """Add the default links between the modem and the device."""
        result_links = await super().async_add_default_links()
        result_notify = await self.async_set_notify_changes()
        return multiple_status(result_links, result_notify)

    def _register_op_flags_and_props(self):
        """Register thermostat operating flags."""
        self._add_property(TEMP_OFFSET, 6, 2, 0, prop_type=PropertyType.ADVANCED)
        self._add_property(HUMIDITY_OFFSET, 7, 3, 0, prop_type=PropertyType.ADVANCED)
        self._add_property(BACKLIGHT, 10, 5, 0)
        self._add_property(CHANGE_DELAY, 11, 6, 0)
        self._add_property(OPERATING_FLAGS, 13, 4, 0, prop_type=PropertyType.ADVANCED)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        op_flags = self._properties[OPERATING_FLAGS]
        self._config[PROGRAM_LOCK_ON] = OpFlagProperty(
            self._address, PROGRAM_LOCK_ON, op_flags, 0
        )
        self._config[KEY_BEEP_ON] = OpFlagProperty(
            self._address, KEY_BEEP_ON, op_flags, 1
        )
        self._config[BUTTON_LOCK_ON] = OpFlagProperty(
            self._address, BUTTON_LOCK_ON, op_flags, 2
        )
        self._config[CELSIUS] = OpFlagProperty(self._address, CELSIUS, op_flags, 3)
        self._config[TIME_24_HOUR_FORMAT] = OpFlagProperty(
            self._address, TIME_24_HOUR_FORMAT, op_flags, 4
        )
        self._config[LED_ON] = OpFlagProperty(self._address, LED_ON, op_flags, 6)

        self._config[CELSIUS].subscribe(self._temp_format_first_set)

    def _register_groups(self):
        """Register the thermostat groups."""
        self._groups[GRP_COOL_ON] = OnOff(COOLING, self._address, GRP_COOL_ON)
        self._groups[GRP_HEAT_ON] = OnOff(HEATING, self._address, GRP_HEAT_ON)
        self._groups[GRP_HUMID_HI_ON] = OnOff(
            DEHUMIDIFYING, self._address, GRP_HUMID_HI_ON
        )
        self._groups[GRP_HUMID_LO_ON] = OnOff(
            HUMIDIFYING, self._address, GRP_HUMID_LO_ON
        )

        self._groups[GRP_TEMP] = Temperature(TEMPERATURE, self._address, GRP_TEMP, 0)
        self._groups[GRP_HUMID] = Humidity(
            HUMIDITY, self._address, group=GRP_HUMID, default=0
        )
        self._groups[GRP_SYS_MODE] = SystemMode(
            SYSTEM_MODE, self._address, group=GRP_SYS_MODE, default=0
        )
        self._groups[GRP_FAN_MODE] = FanMode(
            FAN_MODE, self._address, group=GRP_FAN_MODE, default=4
        )
        self._groups[GRP_COOL_SP] = SetPoint(
            COOL_SET_POINT,
            self._address,
            group=GRP_COOL_SP,
            default=65,
        )
        self._groups[GRP_HEAT_SP] = SetPoint(
            HEAT_SET_POINT,
            self._address,
            group=GRP_HEAT_SP,
            default=95,
        )
        self._groups[GRP_HUMID_HI_SP] = Humidity(
            HUMIDITY_HIGH, self._address, group=GRP_HUMID_HI_SP, default=0
        )
        self._groups[GRP_HUMID_LO_SP] = Humidity(
            HUMIDITY_LOW, self._address, group=GRP_HUMID_LO_SP, default=0
        )

    def _register_handlers_and_managers(self):
        """Register thermostat handlers and managers."""
        super()._register_handlers_and_managers()
        self._managers[GRP_COOL_ON] = OnLevelManager(self._address, GRP_COOL_ON, 0x00)
        self._managers[GRP_HEAT_ON] = OnLevelManager(self._address, GRP_HEAT_ON, 0x00)
        self._managers[GRP_HUMID_HI_ON] = OnLevelManager(
            self._address, GRP_HUMID_HI_ON, 0x00
        )
        self._managers[GRP_HUMID_LO_ON] = OnLevelManager(
            self._address, GRP_HUMID_LO_ON, 0x00
        )

        self._managers[STATUS_COMMAND] = GetThermostatStatus(self._address)
        self._handlers["cool_set_point_handler"] = ThermostatCoolSetPointHandler(
            self._address
        )
        self._handlers["heat_set_point_handler"] = ThermostatHeatSetPointHandler(
            self._address
        )
        self._handlers["humidity_handler"] = ThermostatHumidityHandler(self._address)
        self._handlers["temperature_handler"] = ThermostatTemperatureHandler(
            self._address
        )
        self._handlers["mode_handler"] = ThermostatModeHandler(self._address)

        self._handlers["cool_set_point_command"] = ThermostatCoolSetPointCommand(
            self._address
        )
        self._handlers["heat_set_point_command"] = ThermostatHeatSetPointCommand(
            self._address
        )
        self._handlers["mode_command"] = ThermostatModeCommand(self._address)
        self._handlers["notify_changes_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x08
        )
        self._handlers["humidity_high_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x0B
        )
        self._handlers["humidity_low_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x0C
        )
        self._handlers["set_master"] = ExtendedSetCommand(self._address, 0x00, 0x09)

        self._handlers["op_flag_write"] = ExtendedSetCommand(self._address, 0x00, 0x04)

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe to handlers and managers."""
        super()._subscribe_to_handelers_and_managers()
        self._managers[GRP_COOL_ON].subscribe(self._groups[GRP_COOL_ON].set_value)
        self._managers[GRP_HEAT_ON].subscribe(self._groups[GRP_HEAT_ON].set_value)
        self._managers[GRP_HUMID_HI_ON].subscribe(
            self._groups[GRP_HUMID_HI_ON].set_value
        )
        self._managers[GRP_HUMID_LO_ON].subscribe(
            self._groups[GRP_HUMID_LO_ON].set_value
        )

        self._managers[STATUS_COMMAND].subscribe_status(self._status_received)
        self._managers[STATUS_COMMAND].subscribe_set_point(self._set_point_received)
        self._handlers["cool_set_point_handler"].subscribe(
            self._groups[GRP_COOL_SP].set_value
        )
        self._handlers["heat_set_point_handler"].subscribe(
            self._groups[GRP_HEAT_SP].set_value
        )
        self._handlers["humidity_handler"].subscribe(self._groups[GRP_HUMID].set_value)
        self._handlers["temperature_handler"].subscribe(self._temp_received)
        self._handlers["mode_handler"].subscribe(self._mode_received)

        self._handlers["cool_set_point_command"].subscribe(self._cool_set_point_set)
        self._handlers["heat_set_point_command"].subscribe(self._heat_set_point_set)
        self._handlers["mode_command"].subscribe(self._mode_set)

    def _register_default_links(self):
        """Register default links."""
        super()._register_default_links()
        link_1 = DefaultLink(
            is_controller=True,
            group=GRP_COOL_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_COOL_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_COOL_ON,
        )
        link_2 = DefaultLink(
            is_controller=True,
            group=GRP_HEAT_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HEAT_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HEAT_ON,
        )
        link_3 = DefaultLink(
            is_controller=True,
            group=GRP_HUMID_HI_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HUMID_HI_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HUMID_HI_ON,
        )
        link_4 = DefaultLink(
            is_controller=True,
            group=GRP_HUMID_LO_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HUMID_LO_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HUMID_LO_ON,
        )
        link_ef = DefaultLink(
            is_controller=True,
            group=GRP_NOTIFY,
            dev_data1=0x03,
            dev_data2=0,
            dev_data3=GRP_NOTIFY,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_NOTIFY,
        )
        self._default_links.append(link_1)
        self._default_links.append(link_2)
        self._default_links.append(link_3)
        self._default_links.append(link_4)
        self._default_links.append(link_ef)

    def _status_received(
        self,
        day,
        hour,
        minute,
        second,
        system_mode,
        fan_mode,
        cool_set_point,
        humidity,
        temperature,
        cooling,
        heating,
        celsius,
        heat_set_point,
    ):
        """Receive the status update."""
        self._groups[GRP_COOL_ON].set_value(cooling)
        self._groups[GRP_HEAT_ON].set_value(heating)
        self._groups[GRP_SYS_MODE].set_value(system_mode)
        self._groups[GRP_FAN_MODE].set_value(fan_mode)
        self._groups[GRP_COOL_SP].set_value(cool_set_point)
        self._groups[GRP_HEAT_SP].set_value(heat_set_point)
        if not self._config[CELSIUS].value:
            temperature = to_fahrenheit(temperature)
        self._groups[GRP_TEMP].set_value(temperature)
        self._groups[GRP_HUMID].set_value(humidity)

    def _set_point_received(
        self,
        humidity_high,
        humidity_low,
        firmwire,
        cool_set_point,
        heat_set_point,
        rf_offset,
    ):
        """Receive set point info."""
        self._groups[GRP_COOL_SP].set_value(cool_set_point)
        self._groups[GRP_HEAT_SP].set_value(heat_set_point)
        self._groups[GRP_HUMID_HI_SP].set_value(humidity_high)
        self._groups[GRP_HUMID_LO_SP].set_value(humidity_low)

    def _mode_received(self, system_mode, fan_mode):
        """Receive current temperature notification."""
        self._groups[GRP_SYS_MODE].set_value(system_mode)
        self._groups[GRP_FAN_MODE].set_value(fan_mode)

    def _mode_set(self, thermostat_mode: ThermostatMode):
        """Update the thermostat mode from a set mode command."""
        if thermostat_mode in [ThermostatMode.FAN_ALWAYS_ON, ThermostatMode.FAN_AUTO]:
            self._groups[GRP_FAN_MODE].set_value(thermostat_mode)
        else:
            self._groups[GRP_SYS_MODE].set_value(thermostat_mode)

    def _temp_received(self, degrees):
        """Receive temperature status update and convert to celsius if needed."""
        self._groups[GRP_TEMP].value = degrees

    async def _async_temp_format_changed(self, name, value):
        """Receive notification that the thermostat has changed to/from C/F."""
        await self.async_status()

    def _temp_format_first_set(self, name, value):
        """Set up the trigger for a temperature format change.

        The first time the format is set, we don't need to do anything. If
        the format changes later, we need to get the status update to change the
        measurements from F to C or vise versa.
        """
        self._config[CELSIUS].unsubscribe(self._temp_format_first_set)
        self._config[CELSIUS].subscribe(self._async_temp_format_changed)

    def _cool_set_point_set(self, degrees, zone, deadband):
        """Cool set point changed."""
        self._groups[GRP_COOL_SP].set_value(degrees)

    def _heat_set_point_set(self, degrees, zone, deadband):
        """Cool set point changed."""
        self._groups[GRP_HEAT_SP].set_value(degrees)


class ClimateControl_WirelessThermostat(BatteryDeviceBase, ClimateControl_Thermostat):
    """Wireless Thermostat device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the Wireless Thermostat class."""
        # pylint: disable=super-with-arguments
        super(ClimateControl_WirelessThermostat, self).__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
        )
        self._aldb = ALDBBattery(
            address=address, mem_addr=0x1FFF, run_command=self._run_on_wake
        )

    async def async_set_humidity_high_set_point(self, humidity):
        """Set the humidity high set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_humidity_high_set_point,
            humidity=humidity,
        )

    async def async_set_humidity_low_set_point(self, humidity):
        """Set the humidity low set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_humidity_low_set_point,
            humidity=humidity,
        )

    async def async_set_master(self, master):
        """Set the thermostat master mode."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_master, master=master
        )

    async def async_set_mode(self, thermostat_mode):
        """Set the thermostat mode."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_mode,
            thermostat_mode=thermostat_mode,
        )

    async def async_set_notify_changes(self):
        """Set the thermostat to notify of changes."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_notify_changes
        )

    async def async_set_cool_set_point(self, temperature):
        """Set the cool set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_cool_set_point,
            temperature=temperature,
        )

    async def async_set_day_time(self):
        """Set the thermostat day of the week and time."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_set_day_time)

    async def async_set_heat_set_point(self, temperature):
        """Set the cool set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_heat_set_point,
            temperature=temperature,
        )

    async def async_add_default_links(self):
        """Add default links to the device."""
        self._run_on_wake(self.async_add_default_links_on_wake)

    async def async_add_default_links_on_wake(self):
        """Add default links to the device when the device wakes up."""
        aldb_write_save = self.aldb.async_write
        aldb_load_save = self.aldb.async_load
        self.aldb.async_write = self.aldb.async_write_on_wake
        self.aldb.async_load = self.aldb.async_load_on_wake

        result_links = ResponseStatus.FAILURE
        result_notify = ResponseStatus.FAILURE

        try:
            result_links = await async_add_default_links(self)
            result_notify = await super().async_set_notify_changes()
        finally:
            self.aldb.async_write = aldb_write_save
            self.aldb.async_load = aldb_load_save

        return multiple_status(result_links, result_notify)



"""Filter out command logging.

This is used by the stdout log so that it does not echo the command but the command is sent to the file log.
"""
from logging import Filter


class NoStdoutFilter(Filter):
    """Filter output to standard out."""

    def __init__(self, prefix):
        """Init the CommandFilter class."""
        super().__init__()
        self.prefix = prefix

    def filter(self, record):
        """Filter out commands."""
        if record.msg[0 : len(self.prefix)] == self.prefix:
            return False
        return True


class StdoutFilter(Filter):
    """Allow output to standard out."""

    def __init__(self, prefix):
        """Init the CommandFilter class."""
        super().__init__()
        self.prefix = prefix

    def filter(self, record):
        """Filter out commands."""
        if record.msg.find("Encoding detection") != -1:
            return False
        if record.msg[0 : len(self.prefix)] == self.prefix:
            record.msg = self.strip_prefix(record.msg)
            return True
        return False

    def strip_prefix(self, msg):
        """Strip the prefix from the log message before writing out."""
        return msg[len(self.prefix) :]


class StripPrefixFilter(Filter):
    """Allow output to standard out but strip prefix if it is present."""

    def __init__(self, prefix):
        """Init the CommandFilter class."""
        super().__init__()
        self.prefix = prefix

    def filter(self, record):
        """Filter out commands."""
        if record.msg.find("Encoding detection") != -1:
            return False
        if record.msg[0 : len(self.prefix)] == self.prefix:
            record.msg = self.strip_prefix(record.msg)
        return True

    def strip_prefix(self, msg):
        """Strip the prefix from the log message before writing out."""
        return msg[len(self.prefix) :]



"""All-Link database for battery oppertated devices."""
from collections import namedtuple
import logging

from ..constants import EngineVersion
from .aldb import ALDB

LoadCommandParams = namedtuple("LoadCommandParams", "mem_addr num_recs refresh")
Command = namedtuple("Command", "action params")

_LOGGER = logging.getLogger(__name__)


class ALDBBattery(ALDB):
    """ALDB for battery opperated devices."""

    def __init__(
        self, address, version=EngineVersion.UNKNOWN, mem_addr=0x0FFF, run_command=None
    ):
        """Init the ALDBBattery class."""
        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._commands = []
        self._run_command = run_command

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
    ):
        """Load the All-Link Database."""
        self._run_command(
            self.async_load_on_wake,
            mem_addr=mem_addr,
            num_recs=num_recs,
            refresh=refresh,
        )
        return True

    async def async_write(self, force=False):
        """Write modified records to the device."""
        self._run_command(self.async_write_on_wake, force=force)
        return 0

    async def async_load_on_wake(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
    ):
        """Load the All-Link Database when the device wakes up."""
        return await super().async_load(
            mem_addr=mem_addr,
            num_recs=num_recs,
            refresh=refresh,
        )

    async def async_write_on_wake(self, force=False):
        """Write modified records to the device when the device wakes up."""
        return await super().async_write(force=force)



"""Base class for a derived property."""

from ..constants import PropertyType
from ..topics import EXTENDED_PROPERTIES_CHANGED
from ..utils import subscribe_topic
from .device_flag import DeviceFlagBase


class DerivedProperty(DeviceFlagBase):
    """Base class for a derived property."""

    def __init__(
        self,
        address,
        name,
        value_type,
        is_reversed=False,
        is_read_only=False,
    ):
        """Init the DerivedProperty class."""
        super().__init__(
            address,
            "derived_property",
            name,
            value_type,
            is_reversed,
            is_read_only,
            prop_type=PropertyType.DERIVED,
        )
        subscribe_topic(
            self._properties_updated,
            f"{self._address.id}.{EXTENDED_PROPERTIES_CHANGED}",
        )

    def _properties_updated(self):
        """Update values based on underlying properties."""
        self._call_subscribers(name=self._name, value=self.value)

    def set_value(self, value):
        """Do nothing.

        Loading is done via the underlying property.
        """



"""Manage saving and restoring devices from JSON file."""
import json
import logging
from os import path
from typing import Dict

import aiofiles

from ..address import Address
from ..aldb.aldb_record import ALDBRecord
from ..constants import EngineVersion
from ..device_types.device_base import Device
from ..x10_address import X10Address
from .device_id_manager import DeviceId
from .utils import create_device

DEVICE_INFO_FILE = "insteon_devices.json"
OLD_DEVICE_INFO_FILE = "insteon_plm_device_info.dat"
_LOGGER = logging.getLogger(__name__)


def aldb_rec_to_dict(rec):
    """Convert an All-Link Database Record to dictionary."""
    return {
        "memory": rec.mem_addr,
        "in_use": rec.is_in_use,
        "controller": rec.is_controller,
        "high_water_mark": rec.is_high_water_mark,
        "bit5": rec.is_bit5_set,
        "bit4": rec.is_bit4_set,
        "group": rec.group,
        "target": rec.target.id,
        "data1": rec.data1,
        "data2": rec.data2,
        "data3": rec.data3,
    }


def _dict_to_device(device_dict):
    address = Address(device_dict.get("address"))
    aldb_status = device_dict.get("aldb_status", 0)
    aldb = device_dict.get("aldb", {})
    cat = device_dict.get("cat")
    subcat = device_dict.get("subcat")
    firmware = device_dict.get("firmware")
    engine_version = device_dict.get("engine_version", 0xFF)
    # EngineVersion.UNKNOWN was 0x03, no is 0xFF
    if engine_version == 0x03:
        engine_version = 0xFF
    operating_flags = device_dict.get("operating_flags", {})
    properties = device_dict.get("properties", {})
    first_mem_addr = device_dict.get("first_mem_addr")
    device_id = DeviceId(address, cat, subcat, firmware)
    read_write_mode = device_dict.get("read_write_mode", 0)
    device = create_device(device_id)
    if device:
        device.engine_version = engine_version
        aldb_records = dict_to_aldb_record(aldb)
        device.aldb.load_saved_records(aldb_status, aldb_records, first_mem_addr)
        device.aldb.read_write_mode = read_write_mode
        for flag in operating_flags:
            value = operating_flags[flag]
            if device.operating_flags.get(flag):
                device.operating_flags[flag].set_value(value)
        for flag in properties:
            value = properties[flag]
            if device.properties.get(flag):
                device.properties[flag].set_value(value)
    return device


def _device_to_dict(device_list):
    """Convert a device to a dictionary."""
    device_dict = []
    for addr in device_list:
        device = device_list.get(addr)
        if not isinstance(device.address, X10Address):
            aldb = {}
            for mem in device.aldb:
                rec = device.aldb[mem]
                aldb[mem] = aldb_rec_to_dict(rec)
            operating_flags = {}
            for flag in device.operating_flags:
                operating_flags[flag] = device.operating_flags[flag].value
            properties = {}
            for flag in device.properties:
                properties[flag] = device.properties[flag].value
            device_info = {
                "address": device.address.id,
                "cat": device.cat,
                "subcat": device.subcat,
                "firmware": device.firmware,
                "engine_version": int(device.engine_version),
                "aldb_status": device.aldb.status.value,
                "aldb": aldb,
                "operating_flags": operating_flags,
                "properties": properties,
                "read_write_mode": device.aldb.read_write_mode,
            }
            try:
                device_info["disable_auto_linking"] = device.aldb.disable_auto_linking
                device_info["monitor_mode"] = device.aldb.monitor_mode
                device_info["auto_led"] = device.aldb.auto_led
                device_info["deadman"] = device.aldb.deadman
            except AttributeError:
                pass
            device_info["first_mem_addr"] = device.aldb.first_mem_addr
            device_dict.append(device_info)
    return device_dict


def dict_to_aldb_record(aldb_dict):
    """Convert a dictionary to an ALDB record."""
    records = {}
    for mem_addr in aldb_dict:
        rec = aldb_dict[mem_addr]
        memory = int(mem_addr)
        control_flags = int(rec.get("control_flags", 0))
        in_use = rec.get("in_use", bool(control_flags & 1 << 7))
        controller = rec.get("controller", bool(control_flags & 1 << 6))
        bit5 = rec.get("bit5", bool(control_flags & 1 << 5))
        bit4 = rec.get("bit4", bool(control_flags & 1 << 4))
        high_water_mark = rec.get("high_water_mark", not bool(control_flags & 1 << 1))
        group = int(rec.get("group", 0))
        target = rec.get("target", rec.get("address", "000000"))
        data1 = int(rec.get("data1", 0))
        data2 = int(rec.get("data2", 0))
        data3 = int(rec.get("data3", 0))
        records[memory] = ALDBRecord(
            memory=memory,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
    return records


def _convert_old_device_dict(old_devices):
    """Convert insteonplm saved device file to new device file."""
    new_devices = []
    for device in old_devices:
        new_device = device
        old_product_key = device.get("firmware")
        old_product_key = device.get("product_key", old_product_key)
        new_device["firmware"] = old_product_key
        old_aldb_status = device.get("aldb_status", 0)
        old_aldb = device.get("aldb", {})
        new_device["aldb_status"] = _convert_old_aldb_status(old_aldb_status)
        new_device["aldb"] = _convert_old_aldb(old_aldb)
        new_devices.append(new_device)
    return new_devices


def _convert_old_aldb_status(old_status) -> int:
    """Convert insteonplm ALDB load status to new ALDB load status.

    Old status values:
        EMPTY = 0
        LOADING = 1
        LOADED = 2
        FAILED = 3
        PARTIAL = 4

    New status values:
        EMPTY = 0
        LOADED = 1
        LOADING = 2
        FAILED = 3
        PARTIAL = 4
    """
    if old_status == 1:
        status = 2
    elif old_status == 2:
        status = 1
    else:
        status = old_status
    return status


def _convert_old_aldb(old_aldb):
    """Convert insteonplm ALDB to new ALDB."""
    new_aldb = {}
    for mem_addr in old_aldb:
        rec = old_aldb[mem_addr]
        control_flags = int(rec.get("control_flags", 0))
        in_use = rec.get("in_use", bool(control_flags & 1 << 7))
        controller = rec.get("controller", bool(control_flags & 1 << 6))
        bit5 = rec.get("bit5", bool(control_flags & 1 << 5))
        bit4 = rec.get("bit4", bool(control_flags & 1 << 4))
        high_water_mark = rec.get("high_water_mark", not bool(control_flags & 1 << 1))
        group = int(rec.get("group", 0))
        target = rec.get("address", "000000")
        data1 = int(rec.get("data1", 0))
        data2 = int(rec.get("data2", 0))
        data3 = int(rec.get("data3", 0))

        new_aldb[mem_addr] = {
            "memory": mem_addr,
            "in_use": in_use,
            "controller": controller,
            "high_water_mark": high_water_mark,
            "bit5": bit5,
            "bit4": bit4,
            "group": group,
            "target": target,
            "data1": data1,
            "data2": data2,
            "data3": data3,
        }
    return new_aldb


class SavedDeviceManager:
    """Manage saving and restoring devices from JSON."""

    def __init__(self, workdir, modem):
        """Init the SavedDeviceManager class."""
        self._workdir = workdir
        self._modem = modem

    async def async_save(self, device_list: dict):
        """Save all devices to the `insteon_devices.json` file for faster loading."""
        device_dict = _device_to_dict(device_list)
        await self._write_saved_devices(device_dict)

    async def async_load(self) -> Dict[Address, Device]:
        """Load devices from the saved device file."""
        saved_devices = await self._read_saved_devices()
        device_list = {}
        for saved_device in saved_devices:
            address = Address(saved_device.get("address"))
            if address != self._modem.address:
                device = _dict_to_device(saved_device)
                if device:
                    device_list[address] = device
                    if (
                        device.engine_version == EngineVersion.UNKNOWN
                        and device.cat != 0x03
                    ):
                        await device.async_get_engine_version()
                    _LOGGER.debug(
                        "Device with id %s added to device list "
                        "from saved device data.",
                        address,
                    )
            else:
                aldb_status = saved_device.get("aldb_status", 0)
                aldb = saved_device.get("aldb", {})
                read_write_mode = saved_device.get("read_write_mode", 0)
                first_mem_ddr = saved_device.get("first_mem_addr")
                aldb_records = dict_to_aldb_record(aldb)
                self._modem.aldb.load_saved_records(
                    aldb_status, aldb_records, first_mem_ddr
                )
                self._modem.aldb.read_write_mode = read_write_mode
        return device_list

    async def _read_saved_devices(self):
        """Load device information from the device info file."""
        _LOGGER.debug("Loading saved device info.")

        saved_devices = []
        if not self._workdir:
            _LOGGER.debug("Really Loading saved device info.")
            return saved_devices

        try:
            device_file = path.join(self._workdir, DEVICE_INFO_FILE)
            async with aiofiles.open(device_file, "r") as afp:
                json_file = ""
                json_file = await afp.read()
            try:
                saved_devices = json.loads(json_file)
            except json.decoder.JSONDecodeError:
                _LOGGER.debug("Loading saved device file failed")
        except FileNotFoundError:
            _LOGGER.debug("Saved device file not found")
            saved_devices = await self._read_old_device_file()
        return saved_devices

    async def _write_saved_devices(self, device_list):
        _LOGGER.debug("Writing %d devices to save file", len(device_list))
        device_file = path.join(self._workdir, DEVICE_INFO_FILE)
        try:
            async with aiofiles.open(device_file, "w") as afp:
                out_json = json.dumps(device_list, indent=2)
                await afp.write(out_json)
                await afp.flush()
        except FileNotFoundError as ex:
            _LOGGER.error("Cannot write to file %s", device_file)
            _LOGGER.error("Exception: %s", str(ex))

    async def _read_old_device_file(self):
        """Load device information from the insteonplm device info file."""
        _LOGGER.debug("Loading insteonplm saved device info.")

        saved_devices = {}
        if not self._workdir:
            _LOGGER.debug("Really Loading saved device info.")
            return saved_devices

        try:
            device_file = path.join(self._workdir, OLD_DEVICE_INFO_FILE)
            async with aiofiles.open(device_file, "r") as afp:
                json_file = ""
                json_file = await afp.read()
            try:
                saved_devices = json.loads(json_file)
            except json.decoder.JSONDecodeError:
                _LOGGER.debug("Loading insteonplm saved device file failed")
        except FileNotFoundError:
            _LOGGER.debug("insteonplm saved device file not found")
            return {}
        return _convert_old_device_dict(saved_devices)



"""Open / Close sensor groups."""

from ..address import Address
from .group_base import GroupBase


class NormallyOpen(GroupBase):
    """Normally open sensor state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: int = None,
        status_type: int = 0,
    ):
        """Init the OnLevel class."""
        super().__init__(
            name, address, group, default, value_type=int, status_type=status_type
        )

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        # Off is a Open state and On is an Closed state
        self.value = 0xFF if on_level else 0


class NormallyClosed(GroupBase):
    """Normally closed sensor state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: int = None,
        status_type: int = 0,
    ):
        """Init the NormallyClosed class."""
        super().__init__(
            name, address, group, default, value_type=int, status_type=status_type
        )

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        # Off is a Closed state and On is an Open state
        self.value = 0 if on_level else 0xFF



"""Thermostat temperature up command."""
import asyncio

from .. import ack_handler
from ...topics import THERMOSTAT_SET_HEAT_SETPOINT
from .direct_command import DirectCommandHandlerBase


class ThermostatHeatSetPointCommand(DirectCommandHandlerBase):
    """Manage an outbound THERMOSTAT_SET_COOL_SETPOINT command to a device."""

    def __init__(self, address):
        """Init the ThermostatHeatSetPointCommand class."""
        super().__init__(topic=THERMOSTAT_SET_HEAT_SETPOINT, address=address)
        self._degrees = None
        self._zone = None
        self._deadband = None

    # pylint: disable=arguments-differ
    async def async_send(self, degrees, zone: int = None, deadband: int = None):
        """Send the THERMOSTAT_SET_HEAT_SETPOINT command async."""
        return await super().async_send(degrees=degrees, zone=zone, deadband=deadband)

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response."""
        if user_data["d1"]:
            self._degrees = user_data["d1"] / 2
            self._zone = cmd2
            self._deadband = user_data["d2"] / 2
        else:
            self._degrees = cmd2 / 2
            self._zone = None
            self._deadband = None
        await super().async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)
        await asyncio.sleep(0.2)
        self._degrees = None
        self._zone = None
        self._deadband = None

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(
            degrees=self._degrees, zone=self._zone, deadband=self._deadband
        )



"""Handle All-Link Cleanup Status Reports."""
from ..topics import ALL_LINK_CLEANUP_STATUS_REPORT
from . import ack_handler, nak_handler
from .inbound_base import InboundHandlerBase


class AllLinkCleanupStatusReport(InboundHandlerBase):
    """Handle All-Link Cleanup Status Reports."""

    def __init__(self):
        """Init AllLinkCleanupStatusReport class."""
        super().__init__(topic=ALL_LINK_CLEANUP_STATUS_REPORT)

    @ack_handler
    def success_handler(self):
        """Handle the All-Link Status Report inbound ACK message."""
        self._call_subscribers(success=True)

    @nak_handler
    def failure_handler(self):
        """Handle the All-Link Status Report inbound NAK message."""
        self._call_subscribers(success=False)



"""All-Link Database for devices with no All-Link Database."""

from typing import List, Tuple

from ..address import Address
from ..constants import ALDBStatus, EngineVersion, ReadWriteMode, ResponseStatus
from .aldb_record import ALDBRecord


class NoALDB:
    """All-Link Database for devices with no All-Link Database."""

    def __init__(self, address: Address, version=EngineVersion.I2, mem_addr=0x0FFF):
        """Do nothing."""
        self._address = address

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return 0

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for addr in {}:
            yield addr

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return {}

    def __setitem__(self, mem_addr, record):
        """Do nothing."""

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ", ".join(f"{k}: {repr(v)}" for k, v in attrs.items())

    def items(self):
        """Return the ALDB items."""
        return {}

    @property
    def address(self):
        """Returnt the status of the device."""
        return self._address

    @property
    def status(self) -> ALDBStatus:
        """Return loaded status."""
        return ALDBStatus.LOADED

    @property
    def version(self) -> EngineVersion:
        """Return the ALDB version."""
        return EngineVersion.UNKNOWN

    @property
    def first_mem_addr(self):
        """Return the expected memory address of the first record."""
        return 0x0FFF

    @property
    def high_water_mark_mem_addr(self) -> int:
        """Return the High Water Mark record memory address."""
        return 0x0FFF

    @property
    def is_loaded(self) -> bool:
        """Return loaded always."""
        return True

    @property
    def pending_changes(self):
        """Return pending changes."""
        return {}

    @property
    def read_write_mode(self) -> ReadWriteMode:
        """Emit the modem read mode."""
        return ReadWriteMode.STANDARD

    @read_write_mode.setter
    def read_write_mode(self, value: ReadWriteMode):
        """Set the modem read mode."""

    def clear(self):
        """Clear all records from the ALDB.

        This does not write to the device.
        """

    def clear_pending(self):
        """Remove pending changes."""

    def get(self, mem_addr, default=None):
        """Get the record at address 'mem_addr'."""
        return None

    def get_responders(self, group):
        """Return all responders to this device for a group."""
        return None

    def update_version(self, version):
        """Update the ALDB version number."""

    def subscribe_status_changed(self, listener):
        """Subscribe to notification of ALDB load status changes."""

    def subscribe_record_changed(self, listener):
        """Subscribe to notification of ALDB record changes."""

    def unsubscribe_status_changed(self, listener):
        """Unsubscribe to notification of ALDB load status changes."""

    def unsubscribe_record_changed(self, listener):
        """Unsubscribe to notification of ALDB record changes."""

    async def async_load(self, *args, **kwargs):
        """Load the All-Link Database."""

    def load_saved_records(self, *args, **kwargs):
        """Load All-Link records from a dictionary of saved records."""

    def add(self, *args, **kwargs):
        """Add an All-Link record.

        This method does not write to the device. To write modifications to the device
        use `write` or `async_write`.
        """

    def remove(self, *args, **kwargs):
        """Remove an All-Link record."""

    def modify(self, *args, **kwargs):
        """Modify an All-Link record."""

    async def async_write(self, *args, **kwargs) -> Tuple[int, int]:
        """Write the dirty records to the device."""
        return ResponseStatus.SUCCESS, 0

    def find(self, *args, **kwargs) -> List[ALDBRecord]:
        """Find all records matching the criteria."""
        for no_recs in ():
            yield no_recs

    def set_load_status(self):
        """Review the ALDB records and identify the load status."""



"""Temperature state."""

from ..address import Address
from ..constants import ThermostatMode
from .group_base import GroupBase


class Temperature(GroupBase):
    """Temperature state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: float = None,
        status_type: int = 0,
    ):
        """Init the Temperature class."""
        super().__init__(
            name, address, group, default, value_type=float, status_type=status_type
        )
        self._is_dimmable = True

    # pylint: disable=arguments-differ
    def set_value(self, temperature):
        """Set the temperature value."""
        self.value = temperature

    @property
    def value(self):
        """Return the temperature value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the temperature value."""
        try:
            self._value = self._type(value) if value is not None else None
        except TypeError as ex:
            raise TypeError(
                f"Error setting value of State {self._name}: Must be of type {self._type.__name__}"
            ) from ex
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            value=self._value,
            group=self._group,
        )


class Humidity(GroupBase):
    """Humidity state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: float = None
    ):
        """Init the Temperature class."""
        super().__init__(name, address, group, default, value_type=float)

    # pylint: disable=arguments-differ
    def set_value(self, humidity):
        """Set the temperature value."""
        self.value = humidity


class SetPoint(GroupBase):
    """SetPont state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: float = None
    ):
        """Init the SetPoint class."""
        super().__init__(name, address, group, default, value_type=float)

    # pylint: disable=arguments-differ
    def set_value(self, degrees):
        """Set the Set Point value."""
        self.value = degrees


class SystemMode(GroupBase):
    """System Mode state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: ThermostatMode = None,
    ):
        """Init the SystemMode class."""
        super().__init__(name, address, group, default, value_type=ThermostatMode)

    # pylint: disable=arguments-differ
    def set_value(self, thermostat_mode: ThermostatMode):
        """Set the system mode value."""
        self.value = thermostat_mode

    @property
    def value(self):
        """Return the group value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            if value > 0x0F:
                value = value >> 4
            self._value = self._type(value) if value is not None else None
        except TypeError as ex:
            raise TypeError(
                f"Error setting value of State {self._name}: Must be of type {self._type.__name__}"
            ) from ex
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            value=self._value,
            group=self._group,
        )


class FanMode(GroupBase):
    """Fan Mode state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: ThermostatMode = None,
    ):
        """Init the SystemMode class."""
        super().__init__(name, address, group, default, value_type=ThermostatMode)

    # pylint: disable=arguments-differ
    def set_value(self, thermostat_mode: ThermostatMode):
        """Set the system mode value."""
        self.value = thermostat_mode

    @property
    def value(self):
        """Return the group value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            if value > 0x0F:
                value = value & 0x0F
            self._value = self._type(value) if value is not None else None
        except TypeError as ex:
            raise TypeError(
                f"Error setting value of State {self._name}: Must be of type {self._type.__name__}"
            ) from ex
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            value=self._value,
            group=self._group,
        )



"""Insteon Modem Base Class."""
from abc import ABCMeta
import asyncio
from asyncio import Transport

from ..config import (
    AUTO_LED,
    DEADMAN,
    DISABLE_AUTO_LINKING,
    MONITOR_MODE,
    get_usable_value,
)
from ..config.modem_config import ModemConfiguration
from ..constants import ResponseStatus
from ..handlers.all_link_cleanup_failure_report import AllLinkCleanupFailureReport
from ..handlers.all_link_cleanup_status_report import AllLinkCleanupStatusReport
from ..handlers.get_im_configuration import GetImConfigurationHandler
from ..handlers.set_im_configuration import SetImConfigurationHandler
from ..utils import multiple_status
from .device_base import Device
from .device_commands import GET_IM_CONFIG_COMMAND, SET_IM_CONFIG_COMMAND


class ModemBase(Device, metaclass=ABCMeta):
    """Base class for insteon Modems (PLM and Hub)."""

    __meta__ = ABCMeta

    def __init__(
        self,
        address="000000",
        cat=0x03,
        subcat=0x00,
        firmware=0x00,
        description="",
        model="",
    ):
        """Init the Modem class."""
        super().__init__(address, cat, subcat, firmware, description, model)
        self._aldb = None
        self._subscribe_topics()
        self._protocol = None
        self._transport = None
        self._io_manager = None

    @property
    def connected(self) -> bool:
        """Return true if the transport is connected."""
        if not self._protocol:
            return False
        return self._protocol.connected

    @property
    def protocol(self):
        """Return the protocol."""
        return self._protocol

    @property
    def disable_auto_linking(self):
        """Return the Disable Auto Linking flag value."""
        return self._config[DISABLE_AUTO_LINKING].value

    @property
    def monitor_mode(self):
        """Return the Monitor Mode flag value."""
        return self._config[MONITOR_MODE].value

    @property
    def auto_led(self):
        """Return the Auto LED flag value."""
        return self._config[AUTO_LED].value

    @property
    def deadman(self):
        """Return the Deadman flag value."""
        return self._config[DEADMAN].value

    @protocol.setter
    def protocol(self, value):
        """Set the protocol."""
        self._protocol = value

    @property
    def transport(self):
        """Return the transport."""
        return self._transport

    @transport.setter
    def transport(self, value):
        """Set the transport."""
        if isinstance(value, Transport):
            self._transport = value

    def close(self):
        """Close the connection to the transport."""
        asyncio.ensure_future(self.async_close())

    async def async_close(self):
        """Close the connection to the transport ascynronously."""
        if self._protocol:
            self._protocol.close()
            wait_time = 0.0001
            while self.connected:
                await asyncio.sleep(wait_time)
                wait_time = min(300, 1.5 * wait_time)

    async def async_get_configuration(self):
        """Get the modem flags."""
        return await self._handlers[GET_IM_CONFIG_COMMAND].async_send()

    async def async_set_configuration(
        self,
        disable_auto_linking: bool = None,
        monitor_mode: bool = None,
        auto_led: bool = None,
        deadman: bool = None,
    ):
        """Set the modem flags."""
        self._config[DISABLE_AUTO_LINKING].new_value = disable_auto_linking
        self._config[MONITOR_MODE].new_value = monitor_mode
        self._config[AUTO_LED].new_value = auto_led
        self._config[DEADMAN].new_value = deadman

        disable_auto_linking = get_usable_value(self._config[DISABLE_AUTO_LINKING])
        monitor_mode = get_usable_value(self._config[MONITOR_MODE])
        auto_led = get_usable_value(self._config[AUTO_LED])
        deadman = get_usable_value(self._config[DEADMAN])

        return await self._handlers[SET_IM_CONFIG_COMMAND].async_send(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )

    async def async_read_config(self, read_aldb: bool = True):
        """Read the modem configuration."""
        result_config = await self.async_get_configuration()
        if read_aldb:
            result_aldb = await self.aldb.async_load()
        else:
            result_aldb = ResponseStatus.SUCCESS
        return multiple_status(result_config, result_aldb)

    async def async_write_config(self):
        """Write the configuration changes to the modem."""
        if (
            self._config[DISABLE_AUTO_LINKING].is_dirty
            or self._config[MONITOR_MODE].is_dirty
            or self._config[AUTO_LED].is_dirty
            or self._config[DEADMAN].is_dirty
        ):
            disable_auto_linking = get_usable_value(self._config[DISABLE_AUTO_LINKING])
            monitor_mode = get_usable_value(self._config[MONITOR_MODE])
            auto_led = get_usable_value(self._config[AUTO_LED])
            deadman = get_usable_value(self._config[DEADMAN])

            return await self._handlers[SET_IM_CONFIG_COMMAND].async_send(
                disable_auto_linking=disable_auto_linking,
                monitor_mode=monitor_mode,
                auto_led=auto_led,
                deadman=deadman,
            )
        return ResponseStatus.SUCCESS

    async def async_get_operating_flags(self, group=None):
        """Read the device operating flags."""
        return await self.async_get_configuration()

    async def async_set_operating_flags(self, group=None, force=False):
        """Write the operating flags to the device."""
        return await self.async_set_configuration(
            self._config[DISABLE_AUTO_LINKING].new_value,
            self._config[MONITOR_MODE].new_value,
            self._config[AUTO_LED].new_value,
            self._config[DEADMAN].new_value,
        )

    async def async_get_extended_properties(self, group=None):
        """Get the device extended properties."""

    def _subscribe_topics(self):
        """Subscribe to modem specific topics."""

    def _update_flags(
        self,
        disable_auto_linking: bool,
        monitor_mode: bool,
        auto_led: bool,
        deadman: bool,
    ):
        self._config[DISABLE_AUTO_LINKING].set_value(disable_auto_linking)
        self._config[MONITOR_MODE].set_value(monitor_mode)
        self._config[AUTO_LED].set_value(auto_led)
        self._config[DEADMAN].set_value(deadman)

    def _register_groups(self):
        """No groups to register for modems."""

    def _register_default_links(self):
        """No default links for modems."""

    def _register_handlers_and_managers(self):
        """Register command handlers for modems."""
        super()._register_handlers_and_managers()
        self._handlers["cleanup_status_report"] = AllLinkCleanupStatusReport()
        self._handlers["cleanup_failure_report"] = AllLinkCleanupFailureReport()
        self._handlers[GET_IM_CONFIG_COMMAND] = GetImConfigurationHandler()
        self._handlers[GET_IM_CONFIG_COMMAND].subscribe(self._update_flags)

        self._handlers[SET_IM_CONFIG_COMMAND] = SetImConfigurationHandler()
        self._handlers[SET_IM_CONFIG_COMMAND].subscribe(self._update_flags)

    def _register_events(self):
        """Register events for modems."""

    def _register_op_flags_and_props(self):
        """Register operating flags for modem."""

    def _register_config(self):
        """Register the configuration items of a modem."""
        super()._register_config()
        self._config[DISABLE_AUTO_LINKING] = ModemConfiguration(
            self._address, DISABLE_AUTO_LINKING
        )
        self._config[MONITOR_MODE] = ModemConfiguration(self._address, MONITOR_MODE)
        self._config[AUTO_LED] = ModemConfiguration(self._address, AUTO_LED)
        self._config[DEADMAN] = ModemConfiguration(self._address, DEADMAN)



"""KeypadLinc command handler to trigger a button scene."""
from .. import ack_handler
from ...topics import EXTENDED_TRIGGER_ALL_LINK
from ..to_device.direct_command import DirectCommandHandlerBase


class TriggerSceneOnCommandHandler(DirectCommandHandlerBase):
    """KeypadLinc command handler to trigger a button scene."""

    def __init__(self, address, group):
        """Init the SetLedCommandHandler class."""
        super().__init__(topic=EXTENDED_TRIGGER_ALL_LINK, address=address, group=group)
        self._on_level = None

    # pylint: disable=arguments-differ
    async def async_send(self, on_level=0xFF, fast: bool = False):
        """Set the LED values of the KPL."""
        self._on_level = on_level
        return await super().async_send(group=self._group, on_level=on_level, fast=fast)

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response.

        Required since the BEEP command uses the same cmd1, cmd2 values.
        """
        if not user_data and not user_data["d1"] == self._group:
            return
        return await super().async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(on_level=self._on_level)
        self._on_level = None



"""Manage outbound ON All-Link Broadcast command to a device."""

from ...topics import ON
from .all_link_cleanup_command import AllLinkCleanupCommandHandlerBase


class OnLevelAllLinkCleanupCommand(AllLinkCleanupCommandHandlerBase):
    """Manage an outbound ON All-Link Broadcast command to a device."""

    def __init__(self, address, group):
        """Init the OnLevelCommand class."""
        super().__init__(topic=ON, address=address, group=group)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the ON command async."""
        # In this case the group number goes in the cmd2 field
        return await super().async_send(on_level=self._group, group=0)



"""Insteon message and command handlers."""
import asyncio
from functools import partial, wraps
from inspect import isawaitable, iscoroutinefunction
import logging

from ..address import Address
from ..constants import MessageFlagType
from ..utils import build_topic, subscribe_topic

TIMEOUT = 3  # Time out between ACK and Direct ACK
_LOGGER = logging.getLogger(__name__)


def _register_handler(
    func,
    presets=None,
    topic=None,
    prefix=None,
    address=None,
    group=None,
    message_type=None,
):
    """Register the function with a topic."""
    presets = presets or {}
    topic = presets.get("topic") or topic
    address = presets.get("address") or address
    get_group = presets.get("group", -1)
    group = get_group if get_group != -1 else group
    message_type = presets.get("message_type") or message_type
    full_topic = build_topic(
        topic=topic,
        prefix=presets.get("prefix"),
        address=address,
        group=group,
        message_type=message_type,
    )
    subscribe_topic(func, full_topic)


def _setup_handler(reg_func, func):
    """Set up the handler function."""

    @wraps(func)
    def _wrapper(self, *args, **kwargs):
        """Wrap a function."""
        if iscoroutinefunction(func) or isawaitable(func):
            return asyncio.create_task(func(self, *args, **kwargs))
        return func(self, *args, **kwargs)

    _wrapper.register_handler = reg_func
    return _wrapper


def inbound_handler(func):
    """Register any inbound message handler."""
    reg_func = partial(
        _register_handler,
    )
    return _setup_handler(reg_func, func)


def ack_handler(func):
    """Register the message ACK handler."""
    reg_func = partial(_register_handler, presets={"prefix": "ack"})
    return _setup_handler(reg_func, func)


def nak_handler(func):
    """Register the message NAK handler."""
    reg_func = partial(_register_handler, presets={"prefix": "nak"})
    return _setup_handler(reg_func, func)


def direct_ack_handler(func):
    """Register the DIRECT_ACK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"group": None, "message_type": MessageFlagType.DIRECT_ACK},
    )
    return _setup_handler(reg_func, func)


def status_handler(func):
    """Register the status response handler."""

    def register_status(instance_func, address):
        # This registers all messages for a device but only triggers on
        # status messages if they return within the TIMEOUT period
        address = Address(address)
        subscribe_topic(instance_func, address.id)

    func.register_status = register_status
    return func


def direct_nak_handler(func):
    """Register the DIRECT_NAK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"group": None, "message_type": MessageFlagType.DIRECT_NAK},
    )
    return _setup_handler(reg_func, func)


def broadcast_handler(func):
    """Register the BROADCAST message handler."""

    def _register_broadcast_handler(
        func, topic, address=None, group=None, message_type=None
    ):
        """Register the function with a topic."""
        broadcast_topic = build_topic(
            topic=topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.BROADCAST,
        )
        all_link_topic = build_topic(
            topic=topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )
        subscribe_topic(func, broadcast_topic)
        subscribe_topic(func, all_link_topic)

    reg_func = partial(_register_broadcast_handler)
    return _setup_handler(reg_func, func)


def all_link_cleanup_ack_handler(func):
    """Register the all_link_cleanup ACK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"message_type": MessageFlagType.ALL_LINK_CLEANUP_ACK},
    )
    return _setup_handler(reg_func, func)


def all_link_cleanup_nak_handler(func):
    """Register the all_link_cleanup NAK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"message_type": MessageFlagType.ALL_LINK_CLEANUP_NAK},
    )
    return _setup_handler(reg_func, func)



"""Device manager."""
import asyncio
import logging

import async_timeout

from ..address import Address
from ..constants import AllLinkMode, DeviceAction
from ..device_types.device_base import Device
from ..device_types.modem_base import ModemBase
from ..device_types.x10_base import X10DeviceBase
from ..managers.saved_devices_manager import SavedDeviceManager
from ..subscriber_base import SubscriberBase
from ..topics import DEVICE_LIST_CHANGED
from ..x10_address import X10Address
from .device_id_manager import DeviceId, DeviceIdManager
from .link_manager import (
    async_cancel_linking_mode,
    async_enter_linking_mode,
    async_unlink_devices,
)
from .utils import create_device, create_x10_device

DEVICE_INFO_FILE = "insteon_devices.json"
_LOGGER = logging.getLogger(__name__)
_DEVICE_LOGGERS = []


class DeviceManager(SubscriberBase):
    """Manages the list of active devices."""

    def __init__(self):
        """Init the DeviceManager class."""
        super().__init__(subscriber_topic=DEVICE_LIST_CHANGED)
        self._devices = {}
        self._modem = None
        self._id_manager = DeviceIdManager()
        self._id_manager.subscribe(self._async_device_identified)
        self._loading_saved_lock = asyncio.Lock()

        self._delay_device_inspection = False
        self._to_be_inspected = []
        self._linked_device = asyncio.Queue()
        self.items = self._devices.items

    def __getitem__(self, address) -> Device:
        """Return a a device from the device address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    def __iter__(self):
        """Return an iterator of device addresses."""
        for address in self._devices:
            yield address

    def __setitem__(self, address, device):
        """Add a device to the device list."""
        if device is None:
            _LOGGER.info("Removing device from INSTEON devices list: %s", address.id)
            if address in self._devices:
                self._devices.pop(address)
                self._call_subscribers(address=address.id, action=DeviceAction.REMOVED)
            return

        _LOGGER.info("Adding device to INSTEON devices list: %s", address.id)
        if not isinstance(device, (Device, DeviceId, X10DeviceBase)):
            raise ValueError("Device must be a DeviceId or a Device type.")

        if isinstance(device, DeviceId):
            device = create_device(device)

        self._devices[device.address] = device
        if isinstance(device, Device):
            self._id_manager.set_device_id(
                device.address, device.cat, device.subcat, device.firmware
            )
        self._call_subscribers(address=device.address.id, action=DeviceAction.ADDED)

    def __len__(self):
        """Return the number of devices."""
        return len(self._devices)

    def get(self, address) -> Device:
        """Return a device from an address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    def pop(self, address):
        """Remove a device from the device list."""
        try:
            address = Address(Address)
        except ValueError:
            address = X10Address(address)
        self._devices.pop(address)

    @property
    def modem(self):
        """Return the Insteon Modem."""
        return self._modem

    @modem.setter
    def modem(self, modem):
        """Set the Insteon Modem."""
        if not isinstance(modem, ModemBase):
            raise ValueError("Must be an Insteon Modem object")
        self._modem = modem
        self._devices[self._modem.address] = self._modem

    @property
    def id_manager(self):
        """Return the ID manager instance."""
        return self._id_manager

    @property
    def delay_inspection(self):
        """Return the status of device inspection after identification.

        When a new device is identified the Device Manager will inspect the device
        for its operating flags, extended properties and All-Link Database automatically
        by default. Setting this property to `False` will delay device inspection until
        `inspect_devices` is called.

        """
        return self._delay_device_inspection

    @delay_inspection.setter
    def delay_inspection(self, value):
        """Set the delay inspection property."""
        self._delay_device_inspection = bool(value)

    @property
    def unknown_devices(self):
        """Return a list of addresses where the device type is unknown."""
        return self._id_manager.unknown_devices

    async def async_inspect_device(self, device: Device):
        """Inspect the properties of the devices."""
        await device.async_read_config()
        await self.modem.aldb.async_load()
        await device.async_add_default_links()

    async def _async_ensure_inspect_devices(self):
        """Insepct the properties of the device who's insepection was delayed earlier."""
        while self._to_be_inspected:
            device = self._to_be_inspected.pop()
            await self.async_inspect_device(device)

    def set_id(self, address: Address, cat: int, subcat: int, firmware: int):
        """Add a device override to identify the device information.

        Typical use is to identify devices that do not respond to an ID Request
        such as a battery opperated device.

        """
        address = Address(address)
        device = self[address]
        if device and device.cat == cat and device.subcat == subcat:
            return
        self._id_manager.set_device_id(address, cat, subcat, firmware)

    async def async_identify_device(self, address: Address):
        """Identify a device.

        The device will be placed into the unknown device list to be identified.

        If the device has already been identified, this method will remove the device
        from the known device list. This is typically used when a `set_id` command has
        been run to create a device override. The `async_reidentify_device` command will
        reset that override and allow normal device identification to run.
        """
        self._devices.pop(Address(address))
        await self._id_manager.async_id_device(address=address, refresh=True)

    async def async_add_device(self, address: Address = None, multiple: bool = False):
        """Add one or more devices.

        This command will place the modem in linking mode.

        If an address is entered, it will attempt to place that device in linking mode.

        If multiple is True, the command will continue to place the modem in linking mode until
        the `async_cancel_add_device` command is executed.

        This command can be canceled at any time with the `async_cancel_add_device` command.
        """

        if address is not None:
            address = Address(address)

        self._delay_device_inspection = multiple
        link_next = True
        while not self._linked_device.empty():
            self._linked_device.get_nowait()
        try:
            while link_next:
                await async_enter_linking_mode(
                    link_mode=AllLinkMode.EITHER, group=0, address=address
                )
                # Make sure we don't try to put the same address in linking mode again if multiple is True
                address = None
                async with async_timeout.timeout(3 * 60):
                    linked_address = await self._linked_device.get()
                    link_next = multiple and linked_address
                    yield linked_address
        except asyncio.TimeoutError:
            pass
        finally:
            await async_cancel_linking_mode()
            self._call_subscribers(address=None, action=DeviceAction.COMPLETED)
            if self._delay_device_inspection:
                self._delay_device_inspection = False
                asyncio.ensure_future(self._async_ensure_inspect_devices())

    async def async_remove_device(
        self,
        address: Address = None,
        force: bool = False,
        remove_all_refs: bool = False,
    ):
        """Remove one or more devices.

        This method will remove all links in the modem were the device is the target and
        all links in the device where the modem is the target.

        If `force` is `True`, the device database will not be changed. Only modem links to the device will be removed.

        If `remove_all_refs` is `True`, all references to the device will be removed from all other devices.
        """

        address = Address(address)
        try:
            device_to_delete = self._devices.pop(address)
        except KeyError:
            force = True

        if remove_all_refs:
            for _, device in self.items():
                if device == self.modem or device.address == address:
                    continue
                await async_unlink_devices(device, address)

        if not force and device_to_delete:
            for rec in device_to_delete.aldb.find(
                target=self.modem.address, in_use=True
            ):
                if rec.group == 0:
                    continue
                device_to_delete.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
            await device_to_delete.aldb.async_write()

            for rec in device_to_delete.aldb.find(
                target=self.modem.address, in_use=True, group=0
            ):
                device_to_delete.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
            await device_to_delete.aldb.async_write()

        for rec in self.modem.aldb.find(target=address, in_use=True):
            self.modem.aldb.modify(mem_addr=rec.mem_addr, in_use=False)

        # Return just the modem results because that is what really matters
        result, _ = await self.modem.aldb.async_write()
        return result

    async def async_cancel_all_linking(self):
        """Cancel the All-Link process.

        This command takes the modem out of linking mode.
        """
        await self._linked_device.put(False)

    def add_x10_device(
        self,
        housecode: str,
        unitcode: int,
        x10_feature: str,
        steps: int = 22,
        max_level: int = 255,
    ):
        """Add an X10 device."""
        device = create_x10_device(housecode, unitcode, x10_feature, steps, max_level)
        if device:
            self[device.address] = device
        return device

    async def async_close(self):
        """Close the device ID listener."""
        self._id_manager.close()

    async def async_load(self, workdir="", id_devices=1, load_modem_aldb=1):
        """Load devices from the `insteon_devices.yaml` file and device overrides.

        Parameters:
            workdir: Directory name to find the `insteon_devices.json` file

            id_devices: Indicate if devices should be identified using ID Request
                0: No devices are identified
                1: Unknown devices are identified
                2: All devices are identified
            (default=1)

            load_modem_aldb: Indicate if the Modem ALDB should be loaded
                0: Do not load
                1: Load if not loaded from save file
                2: Load
            (default=1)

        The Modem ALDB is loaded if `refresh` is True or if the saved file has no devices.

        """
        if workdir:
            async with self._loading_saved_lock:
                saved_devices_manager = SavedDeviceManager(workdir, self.modem)
                devices = await saved_devices_manager.async_load()
                for address in devices:
                    self[address] = devices[address]

        if load_modem_aldb == 0:
            load_modem_aldb = False
        elif load_modem_aldb == 2:
            load_modem_aldb = True
        else:
            load_modem_aldb = not self._modem.aldb.is_loaded

        if load_modem_aldb:
            await self._modem.aldb.async_load()

        for mem_addr in self._modem.aldb:
            rec = self._modem.aldb[mem_addr]
            if rec.is_in_use and rec.target != Address("000000"):
                self._id_manager.append(rec.target)

        if id_devices:
            id_all = id_devices == 2
            await self._id_manager.async_id_devices(refresh=id_all)

    async def async_save(self, workdir):
        """Save devices to a device information file."""
        saved_devices_manager = SavedDeviceManager(workdir, self.modem)
        await saved_devices_manager.async_save(self._devices)

    async def _async_device_identified(
        self, device_id: DeviceId, link_mode: AllLinkMode
    ):
        """Device identified by device ID manager."""
        if self._loading_saved_lock.locked():
            return

        await self._linked_device.put(device_id.address)
        device = self._devices.get(device_id.address)
        if (
            device
            and device.cat == device_id.cat
            and device.subcat == device_id.subcat
            and device.firmware == device_id.firmware
        ):
            return

        if link_mode == AllLinkMode.DELETE:
            _LOGGER.debug("Device %s removed", str(device_id.address))
            self[device_id.address] = None
            return

        if device_id.cat is not None:  # How could this be None?
            if (
                self[device_id.address]
                and device_id.cat == self[device_id.address].cat
                and device_id.subcat == self._devices[device_id.address].subcat
            ):
                return

            device = create_device(device_id)
            self[device_id.address] = device
            if device_id.cat != 0x03:
                if self._delay_device_inspection:
                    self._to_be_inspected.append(device)
                else:
                    await self.async_inspect_device(device)
            _LOGGER.debug("Device %s added", str(device.address))
            _DEVICE_LOGGERS.append(
                logging.getLogger(f"pyinsteon.{device_id.address.id}")
            )

    async def _async_remove_all_device_links(self, address: Address):
        """Remove all ALDB records from the device to the modem.

        This does not remove the modem as a controller to group 0.
        That is removed via All-Linking.
        """
        if self._devices.get(address) is None:
            return
        for rec in self[address].aldb.find(target=self.modem.address, in_use=True):
            if rec.group != 0 or rec.is_controller:  # do not process group 0 responder
                self[address].aldb.modify(mem_addr=rec.mem_addr, in_use=False)
        await self[address].aldb.async_write()



"""Broadcast commands sent from the device.

These messages represent commands sent to the Insteon Modem where
the device is the controller and the modem is the responder in the
All-Link database.

Broadcast commands are device state change notifications from the
device to any responder.
"""



"""Insteon Messages."""
from ...constants import MESSAGE_START_CODE
from ...utils import vars_to_bytes, vars_to_repr, vars_to_string
from .message_definition import MessageDefinition


class MessageBase:
    """Base class for all message classes."""

    def __init__(self, msg_def: MessageDefinition, **kwargs):
        """Initialize the MessageBase class."""
        self._start_code = MESSAGE_START_CODE
        self._message_id = msg_def.message_id
        self._fields = msg_def.fields
        for field in self._fields:
            val = kwargs.get(field.name)
            if val is not None and isinstance(val, field.type):
                val = field.type(val)
            setattr(self, field.name, val)

    def __repr__(self):
        """Emit the message in a debug representation."""
        flds = []
        flds.append(("msg_id", self.message_id))
        for field in self._fields:
            flds.append((field.name, getattr(self, field.name)))
        return vars_to_repr(flds)

    def __bytes__(self):
        """Emit the message bytes."""
        data = []
        data.append(self.start_code)
        data.append(self.message_id)
        for field in self._fields:
            data.append(getattr(self, field.name))
        return vars_to_bytes(data)

    def __str__(self):
        """Emit the message as a string."""
        flds = []
        flds.append(("msg_id", self.message_id))
        for field in self._fields:
            flds.append((field.name, getattr(self, field.name)))
        return vars_to_string(flds)

    @property
    def start_code(self):
        """Return message start code 0x02."""
        return self._start_code

    @property
    def message_id(self):
        """Return the message ID."""
        return self._message_id

    @property
    def fields(self):
        """Return the fields contained in the message."""
        return self._fields



"""Operating flags list."""

from ..constants import PropertyType
from .device_flag import DeviceFlagBase


class OperatingFlag(DeviceFlagBase):
    """Operating flag for a device."""

    def __init__(
        self,
        address,
        name,
        value_type: type,
        is_reversed=False,
        is_read_only=False,
        prop_type=PropertyType.STANDARD,
    ):
        """Init the OperatingFlag class."""
        super().__init__(
            address,
            "operating_flag",
            name,
            value_type,
            is_reversed,
            is_read_only,
            prop_type,
        )



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...topics import EXTENDED_READ_WRITE_ALDB
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class WriteALDBCommandHandler(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the WriteALDBCommandHandler."""
        super().__init__(topic=EXTENDED_READ_WRITE_ALDB, address=address)

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        mem_addr: int,
        controller: bool,
        group: int,
        target: Address,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
        in_use: bool = True,
        high_water_mark: bool = False,
        bit5: int = 0,
        bit4: int = 0,
    ):
        """Send ALDB write message asyncronously."""
        return await super().async_send(
            action=0x02,
            mem_addr=mem_addr,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )



"""Get Device Info command handler."""
from ...address import Address
from ...topics import ENTER_UNLINKING_MODE
from .direct_command import DirectCommandHandlerBase


class EnterUnlinkingModeCommand(DirectCommandHandlerBase):
    """Place a device in linking mode command handler."""

    def __init__(self, address: Address):
        """Init the EnterUnlinkingModeCommand class."""
        super().__init__(topic=ENTER_UNLINKING_MODE, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, group: int = 0, extended: bool = False):
        """Send the ENTER_UNLINKING_MODE request asyncronously."""
        return await super().async_send(group=group, extended=extended)



"""Configuration class for the load button configuration property."""

from __future__ import annotations

from ..address import Address
from .derived_property import DerivedProperty
from .extended_property import ExtendedProperty


class LoadButtonProperty(DerivedProperty):
    """Derived property for the load button property."""

    def __init__(
        self,
        address: Address,
        name: str,
        load_button_prop: ExtendedProperty,
        allowed_values: list,
    ) -> None:
        """Init the LoadButtonProperty class."""
        super().__init__(address, name, list, False, False)
        self._load_button_prop = load_button_prop
        self._allowed_values = allowed_values

    @property
    def value(self) -> int | None:
        """Return the value of the load button from the underlying property."""
        return self._load_button_prop.value

    @property
    def new_value(self) -> int | None:
        """Return the modified value of the property."""
        return self._load_button_prop.new_value

    @new_value.setter
    def new_value(self, value) -> None:
        """Modify the value of the property.

        Set value to one of the device buttons.

        Set `new_value` to `None` to retain the current load button value.
        """

        if value is not None and value not in self._allowed_values:
            raise ValueError(f"Value must be None or one of {self._allowed_values}")
        self._load_button_prop.new_value = value

    @property
    def is_dirty(self) -> bool:
        """Return if the underlying property values have changed."""
        return self._load_button_prop.is_dirty

    @property
    def is_loaded(self) -> bool:
        """Return if the underlying property values are loaded.."""
        return self._load_button_prop.is_loaded



"""Base device object."""

from abc import ABC
from datetime import datetime
import logging
from typing import Dict

from ..address import Address
from ..aldb.aldb import ALDB
from ..config.extended_property import ExtendedProperty
from ..config.operating_flag import OperatingFlag
from ..constants import DeviceCategory, EngineVersion, PropertyType, ResponseStatus
from ..default_link import DefaultLink
from ..device_types.device_commands import STATUS_COMMAND
from ..handlers.to_device.engine_version_request import EngineVersionRequest
from ..handlers.to_device.ping import PingCommand
from ..handlers.to_device.product_data_request import ProductDataRequestCommand
from ..managers.get_set_ext_property_manager import GetSetExtendedPropertyManager
from ..managers.get_set_op_flag_manager import GetSetOperatingFlagsManager
from ..managers.link_manager.default_links import async_add_default_links
from ..managers.status_manager import StatusManager
from ..topics import ENGINE_VERSION
from ..utils import multiple_status, publish_topic

_LOGGER = logging.getLogger(__name__)


class Device(ABC):
    """INSTEON Device Class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the Device class."""
        self._address = Address(address)
        self._cat = DeviceCategory(0xFF) if cat is None else DeviceCategory(int(cat))
        self._subcat = subcat
        if self._subcat is None:
            self._subcat = 0x00
        self._firmware = firmware if firmware is not None else 0x00
        self._description = description
        self._model = model
        self._product_id = None
        self._is_battery = False
        self._engine_version = EngineVersion.UNKNOWN

        self._last_communication_received = datetime(1, 1, 1, 1, 1, 1)
        self._product_data_in_aldb = False
        self._groups = {}
        self._handlers = {}
        self._managers = {}
        self._events = {}

        self._aldb = ALDB(self._address)
        self._default_links = []
        self._operating_flags: Dict[str, OperatingFlag] = {}
        self._properties: Dict[str, ExtendedProperty] = {}
        self._config = {}

        self._op_flags_manager = GetSetOperatingFlagsManager(
            self._address, self._operating_flags
        )
        self._ext_property_manager = GetSetExtendedPropertyManager(self._address)

        self._register_groups()
        self._register_events()
        self._register_op_flags_and_props()
        self._register_handlers_and_managers()
        self._subscribe_to_handelers_and_managers()
        self._register_default_links()
        self._register_config()

    @property
    def address(self):
        """Return the INSTEON device address."""
        return self._address

    @property
    def cat(self):
        """Return the INSTEON device category."""
        return self._cat

    @property
    def subcat(self):
        """Return the INSTEON device subcategory."""
        return self._subcat

    @property
    def firmware(self):
        """Return the INSTEON product key."""
        return self._firmware

    @property
    def description(self):
        """Return the INSTEON device description."""
        return self._description

    @property
    def model(self):
        """Return the INSTEON device model number."""
        return self._model

    @property
    def product_id(self):
        """Return the INSTEON product ID."""
        return self._product_id

    @property
    def id(self):
        """Return the ID of the device."""
        return self._address.id

    @property
    def groups(self):
        """Return the device groups / buttons."""
        return self._groups

    @property
    def events(self):
        """Return the device events."""
        return self._events

    @property
    def aldb(self):
        """Return the device All-Link Database."""
        return self._aldb

    @property
    def operating_flags(self):
        """Return the Operating Flags."""
        return self._operating_flags

    @property
    def properties(self):
        """Return the Extended Properties."""
        return self._properties

    @property
    def configuration(self):
        """Return the configuration properties."""
        return self._config

    @property
    def default_links(self):
        """Return the list of default links."""
        return self._default_links

    @property
    def is_battery(self):
        """Return True if the device is battery operated."""
        return self._is_battery

    @property
    def engine_version(self):
        """Return device engine version."""
        return self._engine_version

    @engine_version.setter
    def engine_version(self, value: EngineVersion):
        """Set the device engine version."""
        try:
            version = EngineVersion(value)
        except ValueError:
            version = EngineVersion.UNKNOWN
        if version in [EngineVersion.I2, EngineVersion.I2CS]:
            self._op_flags_manager.extended_write = True
        self._engine_version = version
        publish_topic(
            f"{self._address.id}.{ENGINE_VERSION}", version=self._engine_version
        )

    async def async_status(self, group=None):
        """Get the status of the device."""
        return ResponseStatus.SUCCESS

    async def async_read_config(self, read_aldb: bool = True):
        """Get all configuration settings.

        This includes:
        - Engine Version
        - Operating flags
        - Extended properties
        - All-Link Database records.
        """
        result_engine = await self.async_get_engine_version()
        result_op_flags = await self.async_read_op_flags()
        result_ext_prop = await self.async_read_ext_properties()
        if read_aldb:
            result_aldb = await self._aldb.async_load()
        else:
            result_aldb = ResponseStatus.SUCCESS
        return multiple_status(
            result_engine, result_ext_prop, result_op_flags, result_aldb
        )

    async def async_write_config(self):
        """Write the device configuration to the physical device."""
        result1 = await self.async_write_op_flags()
        result2 = await self.async_write_ext_properties()
        return multiple_status(result1, result2)

    async def async_read_op_flags(self, group=None):
        """Read the device operating flags."""
        return await self._op_flags_manager.async_read(group=group)

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        return await self._op_flags_manager.async_write()

    async def async_read_ext_properties(self, group=None):
        """Get the device extended properties."""
        return await self._ext_property_manager.async_read(group=group)

    async def async_write_ext_properties(self):
        """Write the extended properties."""
        return await self._ext_property_manager.async_write()

    async def async_read_product_id(self):
        """Get the product ID."""
        return await self._handlers["product_data_cmd"].async_send()

    async def async_add_default_links(self):
        """Add the default links betweent he modem and the device."""
        return await async_add_default_links(self)

    async def async_get_engine_version(self):
        """Read the device engine version."""
        return await self._handlers["engine_version_cmd"].async_send()

    async def async_ping(self):
        """Ping the device."""
        cmd = PingCommand(self.address)
        return await cmd.async_send()

    def _register_groups(self):
        """Add the groups to the device."""

    def _register_handlers_and_managers(self):
        """Add all handlers to the device and register listeners."""
        self._handlers["product_data_cmd"] = ProductDataRequestCommand(self._address)
        self._handlers["engine_version_cmd"] = EngineVersionRequest(self._address)
        self._managers[STATUS_COMMAND] = StatusManager(self._address)

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe groups and events to handlers and managers."""
        self._handlers["product_data_cmd"] = ProductDataRequestCommand(self._address)
        self._handlers["engine_version_cmd"].subscribe(self._engine_version_received)

    def _register_default_links(self):
        """Add default links for linking the device to the modem."""
        link_0 = DefaultLink(
            is_controller=False,
            group=0,
            dev_data1=0,
            dev_data2=0,
            dev_data3=0,
            modem_data1=int(self.cat),
            modem_data2=self.subcat,
            modem_data3=self.firmware,
        )
        self._default_links.append(link_0)

    def _register_events(self):
        """Add events that are triggered when events are fired from the device."""

    def _register_op_flags_and_props(self):
        """Add operating flags to the device."""

    def _register_config(self):
        """Set up the configuration properties."""
        for name, prop in self._operating_flags.items():
            if prop.property_type == PropertyType.STANDARD:
                self._config[name] = prop

        for name, prop in self._properties.items():
            if prop.property_type == PropertyType.STANDARD:
                self._config[name] = prop

    def _add_operating_flag(
        self,
        name,
        group,
        bit,
        set_cmd,
        unset_cmd,
        is_reversed=False,
        prop_type=PropertyType.STANDARD,
    ):
        self._op_flags_manager.check_duplicate(name, group, bit, set_cmd, unset_cmd)
        read_only = set_cmd is None or unset_cmd is None
        value_type = bool if bit is not None else int
        self._operating_flags[name] = OperatingFlag(
            self._address, name, value_type, is_reversed, read_only, prop_type
        )
        self._op_flags_manager.subscribe(name, group, bit, set_cmd, unset_cmd)

    def _add_property(
        self,
        name,
        data_field,
        set_cmd,
        group=1,
        bit=None,
        is_reversed=False,
        prop_type=PropertyType.STANDARD,
    ):
        self._ext_property_manager.check_duplicate(group, data_field, bit, set_cmd)
        self._properties[name] = self._ext_property_manager.create(
            name, group, data_field, bit, set_cmd, is_reversed, prop_type
        )

    def _remove_operating_flag(self, name, group=None):
        self._op_flags_manager.unsubscribe(name)
        self._operating_flags.pop(name)

    def _handle_product_data(self, product_id, cat, subcat):
        """Receive and set the product data information."""
        self._product_id = product_id

    def _engine_version_received(self, engine_version):
        """Receive engine version response."""
        self.engine_version = engine_version

    def _product_data_received(self, product_id, cat, subcat):
        """Receive product data response."""
        self._product_id = product_id
        self._cat = cat
        self._subcat = subcat



"""Configuration itmes for the modem."""

from ..constants import PropertyType
from .device_flag import DeviceFlagBase


class ModemConfiguration(DeviceFlagBase):
    """Configuration itmes for the modem."""

    def __init__(self, address, name):
        """Init the ModemConfiguration properties."""
        super().__init__(
            address, "modem_config", name, bool, prop_type=PropertyType.DERIVED
        )



"""Extended property read manager."""
from typing import Dict

from ..address import Address
from ..config.extended_property import ExtendedProperty
from ..constants import ResponseStatus
from .ext_get_manager import ExtenededGetManager


class ExtendedPropertyReadManager(ExtenededGetManager):
    """Extended property read manager."""

    def __init__(
        self,
        address: Address,
        properties: Dict[int, ExtendedProperty],
        cmd2: int = 0,
        data1_read: int = 0x00,
        data2_read: int = 0x00,
        data3_read: int = None,
        data1_resp: int = 0x00,
        data2_resp: int = 0x01,
        data3_resp: int = None,
    ):
        """Init the ExtendedPropertyReadManager class."""
        super().__init__(
            address=address,
            cmd2=cmd2,
            data1_read=data1_read,
            data2_read=data2_read,
            data3_read=data3_read,
            data1_resp=data1_resp,
            data2_resp=data2_resp,
            data3_resp=data3_resp,
        )
        self._props: Dict[int, ExtendedProperty] = properties

    async def _response_received(self, group, data):
        """Update properties based on the extended get response."""
        for index, prop in self._props.items():
            prop.set_value(data[f"data{index}"])
        await self._response_queue.put(ResponseStatus.SUCCESS)



"""Dimmable ramp rate property."""
from ..utils import ramp_rate_to_seconds, seconds_to_ramp_rate
from .derived_property import DerivedProperty


class RampRateProperty(DerivedProperty):
    """Ramp rate property."""

    def __init__(
        self,
        address,
        name,
        ramp_rate_prop,
    ):
        """Init the RampRateProperty class."""
        super().__init__(address, name, float, False, False)

        self._ramp_rate_prop = ramp_rate_prop

    @property
    def value(self):
        """Return the ramp rate property value in seconds."""
        if self._ramp_rate_prop.value is None:
            return None
        return ramp_rate_to_seconds(self._ramp_rate_prop.value)

    @property
    def new_value(self):
        """Return the ramp rate property in seconds."""
        if self._ramp_rate_prop.new_value is None:
            return None
        return ramp_rate_to_seconds(self._ramp_rate_prop.new_value)

    @new_value.setter
    def new_value(self, value):
        """Set the ramp rate property in seconds."""
        if value is None:
            self._ramp_rate_prop.new_value = None
            return
        value = float(value)
        ramp_rate = seconds_to_ramp_rate(value)
        if int(ramp_rate) == self._ramp_rate_prop.value:
            self._ramp_rate_prop.new_value = None
            return
        self._ramp_rate_prop.new_value = int(ramp_rate)

    @property
    def is_loaded(self):
        """Return if the Operating flag has been loaded."""
        return self._ramp_rate_prop.is_loaded

    @property
    def is_dirty(self):
        """Return if the Operating flag has been changed."""
        return self._ramp_rate_prop.is_dirty



"""Handle an outbound direct message to a device."""

from abc import ABCMeta
import asyncio

import async_timeout

from .. import ack_handler, direct_ack_handler, direct_nak_handler, nak_handler
from ...constants import MessageFlagType, ResponseStatus
from ..outbound_base import OutboundHandlerBase

TIMEOUT = 6  # Wait time for device response


class DirectCommandHandlerBase(OutboundHandlerBase):
    """Abstract base class for outbound direct message handling."""

    __meta__ = ABCMeta

    def __init__(self, topic, address, group=None, message_type=MessageFlagType.DIRECT):
        """Init the DirectCommandHandlerBase class."""
        self._response_lock = asyncio.Lock()
        self._direct_response = asyncio.Queue()
        super().__init__(topic, address=address, group=group, message_type=message_type)

    @property
    def response_lock(self) -> asyncio.Lock:
        """Lock to manage the response between ACK and Direct ACK."""
        return self._response_lock

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        ack_response = await super().async_send(address=self._address, **kwargs)
        if ack_response == ResponseStatus.SUCCESS:
            try:
                async with async_timeout.timeout(TIMEOUT + 0.1):
                    return await self._message_response.get()
            except asyncio.TimeoutError:
                pass
        return ResponseStatus.FAILURE

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle Direct Command ACK message."""
        await self._async_handle_ack()
        try:
            async with async_timeout.timeout(TIMEOUT):
                async with self._response_lock:
                    direct_response = await self._direct_response.get()
                    await self._message_response.put(direct_response)
        except asyncio.TimeoutError:
            await self._message_response.put(ResponseStatus.DEVICE_UNRESPONSIVE)

    @nak_handler
    async def async_handle_nak(self, cmd1, cmd2, user_data):
        """Handle the NAK response from the modem."""
        await self._async_handle_nak()

    @direct_nak_handler
    async def async_handle_direct_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the message ACK."""
        await asyncio.sleep(0.05)
        try:
            response = ResponseStatus(cmd2)
        except ValueError:
            response = ResponseStatus.FAILURE
        if self._response_lock.locked():
            await self._direct_response.put(response)
            await asyncio.sleep(0.05)
            self._update_subscribers_on_direct_nak(
                cmd1, cmd2, target, user_data, hops_left
            )
            await asyncio.sleep(0.05)

    @direct_ack_handler
    async def async_handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK."""
        # Need to make sure the ACK has time to aquire the lock
        await asyncio.sleep(0.05)
        if self._response_lock.locked():
            await self._direct_response.put(ResponseStatus.SUCCESS)
            self._update_subscribers_on_direct_ack(
                cmd1, cmd2, target, user_data, hops_left
            )
            await asyncio.sleep(0.05)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""

    def _update_subscribers_on_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers on NAK received."""

    def _update_subscribers_on_direct_nak(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers on DIRECT NAK received."""



"""Receive an All-Link record for an Insteon Modem."""

import logging

from ..address import Address
from ..constants import AllLinkMode
from ..data_types.all_link_record_flags import AllLinkRecordFlags
from ..topics import READ_EEPROM_RESPONSE
from . import inbound_handler
from .inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ReadEepromResponseHandler(InboundHandlerBase):
    """Receive a READ EEPROM response for an Insteon Modem."""

    def __init__(self):
        """Init the AllLinkRecordResponse class."""
        super().__init__(READ_EEPROM_RESPONSE)
        self._has_subscriber = False

    def subscribe(self, callback, force_strong_ref=False):
        """Subscribe listeners to the topic.

        We only want one modem to be subsribed to this method. The first one
        will always be the one we want subscribed so reject the others.
        """
        if self._has_subscriber:
            return
        self._has_subscriber = True
        super().subscribe(callback, force_strong_ref=force_strong_ref)

    @inbound_handler
    def receive_record(
        self,
        mem_addr: int,
        flags: AllLinkRecordFlags,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
    ):
        """Recieve an all link record."""
        controller = flags.link_mode == AllLinkMode.CONTROLLER
        self._call_subscribers(
            mem_addr=mem_addr,
            in_use=flags.is_in_use,
            high_water_mark=flags.is_hwm,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=flags.is_bit_5_set,
            bit4=flags.is_bit_4_set,
        )



"""Sensor/Actuator devices."""

import asyncio

from ..config import (
    DELAY,
    LED_BLINK_ON_TX_ON,
    MOMENTARY_DELAY,
    MOMENTARY_FOLLOW_SENSE,
    MOMENTARY_MODE_ON,
    MOMENTARY_ON_OFF_TRIGGER,
    PRESCALER,
    PROGRAM_LOCK_ON,
    RELAY_MODE,
    RELAY_ON_SENSE_ON,
    SENSE_SENDS_OFF,
    X10_HOUSE,
    X10_OFF,
    X10_UNIT,
)
from ..config.momentary_delay import MomentaryDelayProperty
from ..config.relay_mode import RelayModeProperty
from ..constants import PropertyType, RelayMode
from ..default_link import DefaultLink
from ..events import CLOSE_EVENT, OFF_EVENT, ON_EVENT, OPEN_EVENT, Event
from ..groups import OPEN_CLOSE_SENSOR, RELAY
from ..groups.on_off import OnOff
from ..groups.open_close import NormallyClosed
from ..handlers.to_device.off import OffCommand
from ..handlers.to_device.on_level import OnLevelCommand
from ..managers.on_level_manager import OnLevelManager
from .device_base import Device
from .device_commands import OFF_COMMAND, ON_COMMAND, STATUS_COMMAND
from .on_off_responder_base import OnOffResponderBase

ON_LEVEL_MANAGER = "on_level_manager"
RELAY_GROUP = 1
SENSOR_GROUP = 2


class SensorsActuators(OnOffResponderBase):
    """Sensor Actuator device."""


class SensorsActuators_IOLink(Device):
    """I/O Link device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the SensorsActuators_IOLink class."""
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )
        self._op_flags_manager.extended_write = True
        self._status_lock = asyncio.Lock()

    @property
    def relay_mode(self) -> RelayMode:
        """Return the relay mode."""
        if not self._operating_flags[MOMENTARY_MODE_ON].value:
            return RelayMode.LATCHING
        if self._operating_flags[MOMENTARY_FOLLOW_SENSE].value:
            return RelayMode.MOMENTARY_C
        if self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].value:
            return RelayMode.MOMENTARY_B
        return RelayMode.MOMENTARY_A

    async def async_set_relay_mode(self, relay_mode: RelayMode):
        """Set the relay mode and write to device."""

        def set_relay_mode(momentary_mode_on, momentary_follow_sense, momentary_on_off):
            """Set the values of the underlying properties."""
            self._operating_flags[MOMENTARY_MODE_ON].new_value = momentary_mode_on
            self._operating_flags[
                MOMENTARY_FOLLOW_SENSE
            ].new_value = momentary_follow_sense
            self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].new_value = momentary_on_off

        if relay_mode is None:
            set_relay_mode(None, None, None)
            return

        relay_mode = RelayMode(int(relay_mode))

        if relay_mode == RelayMode.LATCHING:
            set_relay_mode(False, False, False)
        elif relay_mode == RelayMode.MOMENTARY_A:
            set_relay_mode(True, False, False)
        elif relay_mode == RelayMode.MOMENTARY_B:
            set_relay_mode(True, False, True)
        elif relay_mode == RelayMode.MOMENTARY_C:
            set_relay_mode(True, True, False)

        await self.async_write_op_flags()

    async def async_set_momentary_delay(self, seconds):
        """Set the delay in seconds.

        When the IOLinc is in momentary mode (A, B or C) the relay will
        turn off after the delay period.
        """
        delay = seconds * 10
        prescaler = 1
        if delay > 255:
            prescaler = int(round(delay / 255 + 0.5, 0))
            delay = int(round(delay / prescaler, 0))
        self._properties[PRESCALER].new_value = prescaler
        self._properties[DELAY].new_value = delay
        await self.async_write_ext_properties()

    def on(self, group: int = 0):
        """Turn on the relay."""
        self._handlers[RELAY_GROUP][ON_COMMAND].send()

    async def async_on(self, group: int = 0):
        """Turn on the relay."""
        return await self._handlers[RELAY_GROUP][ON_COMMAND].async_send()

    def off(self, group: int = 0):
        """Turn off the relay."""
        self._handlers[RELAY_GROUP][OFF_COMMAND].send()

    async def async_off(self, group: int = 0):
        """Turn off the relay."""
        return await self._handlers[RELAY_GROUP][OFF_COMMAND].async_send()

    async def async_status(self, group=None):
        """Get the device status."""
        status_type = self._groups[group].status_type if group is not None else None
        return await self._managers[STATUS_COMMAND].async_status(status_type)

    def _register_op_flags_and_props(self):
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(
            RELAY_ON_SENSE_ON, 0, 2, 4, 5, prop_type=PropertyType.ADVANCED
        )  # Sensor triggers relay
        self._add_operating_flag(
            MOMENTARY_MODE_ON, 0, 3, 6, 7, prop_type=PropertyType.ADVANCED
        )
        self._add_operating_flag(
            MOMENTARY_ON_OFF_TRIGGER, 0, 4, 0x12, 0x13, prop_type=PropertyType.ADVANCED
        )
        self._add_operating_flag(X10_OFF, 0, 5, 0x0C, 0x0D)
        self._add_operating_flag(SENSE_SENDS_OFF, 0, 6, 0x0E, 0x0F)
        self._add_operating_flag(
            MOMENTARY_FOLLOW_SENSE, 0, 7, 0x14, 0x15, prop_type=PropertyType.ADVANCED
        )  # Check sensor before triggering?

        self._add_property(PRESCALER, 3, 7, prop_type=PropertyType.ADVANCED)
        self._add_property(DELAY, 4, 6, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[MOMENTARY_DELAY] = MomentaryDelayProperty(
            self._address,
            MOMENTARY_DELAY,
            self._properties[DELAY],
            self._properties[PRESCALER],
        )
        self._config[RELAY_MODE] = RelayModeProperty(
            self._address,
            RELAY_MODE,
            self._operating_flags[MOMENTARY_MODE_ON],
            self._operating_flags[MOMENTARY_FOLLOW_SENSE],
            self._operating_flags[MOMENTARY_ON_OFF_TRIGGER],
        )

    def _register_handlers_and_managers(self):
        """Register handlers and managers.

        The relay is controlled through group 1.
        The sensor and the relay both use group 1 to notify of changes. T
        When an on/off broadcast message is received, we need to check what the
        current status of the device is for both the relay and the sensor.
        """
        super()._register_handlers_and_managers()

        self._managers[ON_LEVEL_MANAGER] = OnLevelManager(self._address, 1)

        self._handlers[RELAY_GROUP] = {}
        self._handlers[RELAY_GROUP][ON_COMMAND] = OnLevelCommand(
            self._address, RELAY_GROUP
        )
        self._handlers[RELAY_GROUP][OFF_COMMAND] = OffCommand(
            self._address, RELAY_GROUP
        )

        self._handlers[SENSOR_GROUP] = {}

    def _register_groups(self):
        """Register the device groups.

        There really is only one group in the device, group 1. Both the relay and the sensor
        both send updates on group 1 for broadcast messages. We create two groups because
        we can keep the status of the relay and the sensor separate.
        """
        self._groups[RELAY_GROUP] = OnOff(
            RELAY, self._address, RELAY_GROUP, status_type=0
        )
        self._groups[SENSOR_GROUP] = NormallyClosed(
            OPEN_CLOSE_SENSOR, self._address, SENSOR_GROUP, status_type=1
        )

    def _register_events(self):
        self._events[ON_EVENT] = Event(ON_EVENT, self._address, 0)
        self._events[OFF_EVENT] = Event(OFF_EVENT, self._address, 0)
        self._events[RELAY_GROUP] = {}
        self._events[RELAY_GROUP][ON_EVENT] = Event(
            ON_EVENT, self._address, RELAY_GROUP
        )
        self._events[RELAY_GROUP][OFF_EVENT] = Event(
            OFF_EVENT, self._address, RELAY_GROUP
        )
        self._events[SENSOR_GROUP] = {}
        self._events[SENSOR_GROUP][OPEN_EVENT] = Event(
            OPEN_EVENT, self._address, SENSOR_GROUP
        )
        self._events[SENSOR_GROUP][CLOSE_EVENT] = Event(
            CLOSE_EVENT, self._address, SENSOR_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe to handelers and managers.

        The relay is controlled through group 1.
        The sensor and the relay both use group 1 to notify of changes. T
        When an on/off broadcast message is received, we need to check what the
        current status of the device is for both the relay and the sensor.
        """
        super()._subscribe_to_handelers_and_managers()
        switch_on_event = self._events[RELAY_GROUP][ON_EVENT]
        switch_off_event = self._events[RELAY_GROUP][OFF_EVENT]
        on_cmd = self._handlers[RELAY_GROUP][ON_COMMAND]
        off_cmd = self._handlers[RELAY_GROUP][OFF_COMMAND]

        on_cmd.subscribe(self._async_switch_changed)
        off_cmd.subscribe(self._async_switch_changed)
        on_cmd.subscribe(switch_on_event.trigger)
        off_cmd.subscribe(switch_off_event.trigger)

        self._managers[ON_LEVEL_MANAGER].subscribe(self._async_on_off_received)
        self._managers[STATUS_COMMAND].add_status_type(0, self._handle_switch_status)
        self._managers[STATUS_COMMAND].add_status_type(1, self._handle_sensor_status)

    def _register_default_links(self):
        link = DefaultLink(
            is_controller=True,
            group=RELAY_GROUP,
            dev_data1=0,
            dev_data2=0,
            dev_data3=0,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link)

    async def _async_switch_changed(self, on_level):
        """Catch on/off signal and fire appropriate response."""
        self._groups[RELAY_GROUP].value = on_level
        if on_level and self._operating_flags[MOMENTARY_MODE_ON].value:
            await self._delay_wait()

    def _calc_delay(self):
        """Calculate the momentary delay based on properties and flags."""
        delay = self._properties[DELAY].value
        delay = 255 if delay == 0 else delay
        if self._properties[PRESCALER].value is not None:
            prescaler = max(self._properties[PRESCALER].value, 1)
        else:
            prescaler = 1
        return delay * prescaler / 10

    async def _delay_wait(self):
        await asyncio.sleep(self._calc_delay())
        self._groups[RELAY_GROUP].value = 0
        self._events[RELAY_GROUP][OFF_EVENT].trigger(on_level=0)

    def _handle_switch_status(self, db_version, status):
        """Handle status response."""
        self._groups[RELAY_GROUP].set_value(status)

    def _handle_sensor_status(self, db_version, status):
        """Handle status response."""
        self._groups[SENSOR_GROUP].set_value(status)

    async def _async_on_off_received(self, on_level):
        """Process an on or off broadcast command."""
        if on_level:
            self._events[ON_EVENT].trigger(on_level=255)
        else:
            self._events[OFF_EVENT].trigger(on_level=0)
        orig_sensor = self._groups[SENSOR_GROUP].value
        orig_relay = self._groups[RELAY_GROUP].value

        # Need to get status since we don't know if the on/off message was for the relay
        # or the sensor. Get the sensor first since that is the most likely change.
        await self.async_status(group=1)
        if self._groups[SENSOR_GROUP].value != orig_sensor:
            if self._groups[SENSOR_GROUP].value:
                self._events[SENSOR_GROUP][OPEN_EVENT].trigger(on_level=255)
            else:
                self._events[SENSOR_GROUP][CLOSE_EVENT].trigger(on_level=0)

        await self.async_status(group=0)
        if self._groups[RELAY_GROUP].value != orig_relay:
            if self._groups[RELAY_GROUP].value:
                self._events[RELAY_GROUP][ON_EVENT].trigger(on_level=255)
                if self._operating_flags[MOMENTARY_MODE_ON].value:
                    await self._delay_wait()
            else:
                self._events[RELAY_GROUP][OFF_EVENT].trigger(on_level=0)



"""Convert a message to a topic and an args, kwargs arguments."""
import logging
from typing import Any, Dict, Tuple

from ..commands import commands
from ..constants import MessageFlagType
from ..topics import (
    ALL_LINK_CLEANUP_FAILURE_REPORT,
    ALL_LINK_CLEANUP_STATUS_REPORT,
    ALL_LINK_RECORD_RESPONSE,
    ALL_LINKING_COMPLETED,
    BUTTON_EVENT_REPORT,
    CANCEL_ALL_LINKING,
    GET_ALL_LINK_RECORD_FOR_SENDER,
    GET_FIRST_ALL_LINK_RECORD,
    GET_IM_CONFIGURATION,
    GET_IM_INFO,
    GET_NEXT_ALL_LINK_RECORD,
    LED_OFF,
    LED_ON,
    MANAGE_ALL_LINK_RECORD,
    READ_EEPROM,
    READ_EEPROM_RESPONSE,
    RESET_IM,
    RF_SLEEP,
    SEND_ALL_LINK_COMMAND,
    SET_ACK_MESSAGE_BYTE,
    SET_ACK_MESSAGE_TWO_BYTES,
    SET_HOST_DEVICE_CATEGORY,
    SET_IM_CONFIGURATION,
    SET_NAK_MESSAGE_BYTE,
    START_ALL_LINKING,
    STATUS_REQUEST,
    USER_RESET_DETECTED,
    WRITE_EEPROM,
    X10_RECEIVED,
    X10_SEND,
)
from ..utils import build_topic
from .messages.inbound import Inbound

MSG_CONVERTER = {}
_LOGGER = logging.getLogger(__name__)


def convert_to_topic(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Convert a message to a topic defintion."""
    converter = MSG_CONVERTER[msg.message_id]
    return converter(msg)


def _get_group_from_msg(topic, message_type, target, cmd2, user_data):
    """Derive the group number of the message from the message type.

    message_flag_type: MessageFlagType 0 to 7:
        DIRECT = 0: Group number is 1 unless extended then group number in d1
        DIRECT_ACK = 1: Group not applicable
        ALL_LINK_CLEANUP = 2: Group number is cmd2
        ALL_LINK_CLEANUP_ACK = 3: Not applicable
        BROADCAST = 4: Group number is in the lower byte of the target field
        DIRECT_NAK = 5: Not applicable
        ALL_LINK_BROADCAST = 6: Not applicable
        ALL_LINK_CLEANUP_NAK = 7: Not applicable
    """
    if topic == STATUS_REQUEST and message_type == MessageFlagType.DIRECT:
        return cmd2

    if message_type == MessageFlagType.DIRECT:
        if user_data:
            return user_data["d1"]
        return 1
    if message_type == MessageFlagType.ALL_LINK_CLEANUP:
        return cmd2
    if message_type in [MessageFlagType.BROADCAST, MessageFlagType.ALL_LINK_BROADCAST]:
        if not target:
            return None
        return target.low
    return None


def _create_rcv_std_ext_msg(topic, address, flags, cmd1, cmd2, target, user_data):
    if commands.use_group(topic):
        group = _get_group_from_msg(topic, flags.message_type, target, cmd2, user_data)
    else:
        group = None
    topic = build_topic(
        topic=topic,
        prefix=None,
        address=address,
        group=group,
        message_type=flags.message_type,
    )
    hops_left = flags.hops_left
    kwargs = {
        "cmd1": cmd1,
        "cmd2": cmd2,
        "target": target,
        "user_data": user_data,
        "hops_left": hops_left,
    }
    return (topic, kwargs)


def standard_received(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from a STANDARD_RECEIVED message."""
    for topic in commands.get_topics(msg.cmd1, msg.cmd2, msg.flags, None):
        yield _create_rcv_std_ext_msg(
            topic, msg.address, msg.flags, msg.cmd1, msg.cmd2, msg.target, None
        )


def extended_received(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from a EXTENDED_RECEIVED message."""
    for topic in commands.get_topics(msg.cmd1, msg.cmd2, msg.flags, msg.user_data):
        yield _create_rcv_std_ext_msg(
            topic, msg.address, msg.flags, msg.cmd1, msg.cmd2, msg.target, msg.user_data
        )


def x10_received(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an X10_RECEIVED message."""
    topic = X10_RECEIVED
    kwargs = {"raw_x10": msg.raw_x10, "x10_flag": msg.x10_flag}
    yield (topic, kwargs)


def all_linking_completed(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an ALL_LINKING_COMPLETED message."""
    topic = ALL_LINKING_COMPLETED
    kwargs = {
        "link_mode": msg.link_mode,
        "group": msg.group,
        "target": msg.target,
        "cat": msg.cat,
        "subcat": msg.subcat,
        "firmware": msg.firmware,
    }
    yield (topic, kwargs)


def button_event_report(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from a BUTTON_EVENT_REPORT message."""
    topic = BUTTON_EVENT_REPORT
    kwargs = {"event": msg.event}
    yield (topic, kwargs)


def user_reset_detected(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from a USER_RESET_DETECTED message."""
    topic = USER_RESET_DETECTED
    kwargs = {}
    yield (topic, kwargs)


def all_link_cleanup_failure_report(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an ALL_LINK_CLEANUP_FAILURE_REPORT message."""
    topic = ALL_LINK_CLEANUP_FAILURE_REPORT
    kwargs = {"error": msg.error, "group": msg.group, "target": msg.target}
    yield (topic, kwargs)


def all_link_record_response(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an ALL_LINK_RECORD_RESPONSE message."""
    topic = ALL_LINK_RECORD_RESPONSE
    kwargs = {
        "flags": msg.flags,
        "group": msg.group,
        "target": msg.target,
        "data1": msg.data1,
        "data2": msg.data2,
        "data3": msg.data3,
    }
    yield (topic, kwargs)


def all_link_cleanup_status_report(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an ALL_LINK_CLEANUP_STATUS_REPORT message."""
    topic = build_topic(ALL_LINK_CLEANUP_STATUS_REPORT, prefix=msg.ack)
    kwargs = {}
    yield (topic, kwargs)


def read_eeprom_response(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an read_eeprom message."""
    topic = build_topic(topic=READ_EEPROM_RESPONSE)
    mem_addr = (msg.mem_hi << 8) + msg.mem_low + 7

    kwargs = {
        "mem_addr": mem_addr,
        "flags": msg.flags,
        "group": msg.group,
        "target": msg.target,
        "data1": msg.data1,
        "data2": msg.data2,
        "data3": msg.data3,
    }
    yield (topic, kwargs)


def get_im_info(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an GET_IM_INFO message."""
    topic = build_topic(prefix=msg.ack, topic=GET_IM_INFO)
    if len(msg) == 3:
        kwargs = {}
    else:
        kwargs = {
            "address": msg.address,
            "cat": msg.cat,
            "subcat": msg.subcat,
            "firmware": msg.firmware,
        }
    yield (topic, kwargs)


def send_all_link_command(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an SEND_ALL_LINK_COMMAND message."""
    topic = build_topic(prefix=msg.ack, topic=SEND_ALL_LINK_COMMAND)
    kwargs = {"group": msg.group, "cmd1": msg.cmd1, "cmd2": msg.cmd2}
    yield (topic, kwargs)


def _create_send_std_ext(topic, address, flags, cmd1, cmd2, user_data, ack):
    if commands.use_group(topic):
        group = _get_group_from_msg(topic, flags.message_type, None, cmd2, user_data)
    else:
        group = None

    topic = build_topic(
        topic=topic,
        prefix=str(ack),
        address=address,
        group=group,
        message_type=flags.message_type,
    )
    kwargs = {"cmd1": cmd1, "cmd2": cmd2, "user_data": user_data}
    return (topic, kwargs)


def send_standard_or_extended_message(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Convert standard and extended messages to topic."""
    def_topic = "send_extended" if msg.flags.is_extended else "send_standard"
    user_data = msg.user_data if msg.flags.is_extended else None

    found_topic = False
    user_data = None if not hasattr(msg, "user_data") else msg.user_data
    for topic in commands.get_topics(msg.cmd1, msg.cmd2, msg.flags, user_data):
        found_topic = True
        yield _create_send_std_ext(
            topic, msg.address, msg.flags, msg.cmd1, msg.cmd2, user_data, msg.ack
        )
    if not found_topic:
        yield _create_send_std_ext(
            def_topic, msg.address, msg.flags, msg.cmd1, msg.cmd2, user_data, msg.ack
        )


def x10_send(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an X10_SEND message."""
    topic = build_topic(prefix=msg.ack, topic=X10_SEND)
    kwargs = {"raw_x10": msg.raw_x10, "x10_flag": msg.x10_flag}
    yield (topic, kwargs)


def start_all_linking(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an start_all_linking message."""
    topic = build_topic(prefix=msg.ack, topic=START_ALL_LINKING)
    kwargs = {"link_mode": msg.link_mode, "group": msg.group}
    yield (topic, kwargs)


def cancel_all_linking(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an cancel_all_linking message."""
    topic = build_topic(prefix=msg.ack, topic=CANCEL_ALL_LINKING)
    kwargs = {}
    yield (topic, kwargs)


def set_host_device_category(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an set_host_device_category message."""
    topic = build_topic(prefix=msg.ack, topic=SET_HOST_DEVICE_CATEGORY)
    kwargs = {"cat": msg.cat, "subcat": msg.subcat, "firmware": msg.firmware}
    yield (topic, kwargs)


def reset_im(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an reset_im message."""
    topic = build_topic(prefix=msg.ack, topic=RESET_IM)
    kwargs = {}
    yield (topic, kwargs)


def set_ack_message_byte(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an set_ack_message_byte message."""
    topic = build_topic(prefix=msg.ack, topic=SET_ACK_MESSAGE_BYTE)
    kwargs = {"cmd2": msg.cmd2}
    yield (topic, kwargs)


def get_first_all_link_record(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an get_first_all_link_record message."""
    topic = build_topic(prefix=msg.ack, topic=GET_FIRST_ALL_LINK_RECORD)
    kwargs = {}
    yield (topic, kwargs)


def get_next_all_link_record(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an get_next_all_link_record message."""
    topic = build_topic(prefix=msg.ack, topic=GET_NEXT_ALL_LINK_RECORD)
    kwargs = {}
    yield (topic, kwargs)


def set_im_configuration(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an set_im_configuration message."""
    topic = build_topic(prefix=msg.ack, topic=SET_IM_CONFIGURATION)

    kwargs = {
        "disable_auto_linking": msg.flags.is_auto_link,
        "monitor_mode": msg.flags.is_monitor_mode,
        "auto_led": msg.flags.is_auto_led,
        "deadman": msg.flags.is_disable_deadman,
    }
    yield (topic, kwargs)


def get_all_link_record_for_sender(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an get_all_link_record_for_sender message."""
    topic = build_topic(prefix=msg.ack, topic=GET_ALL_LINK_RECORD_FOR_SENDER)
    kwargs = {}
    yield (topic, kwargs)


def led_on(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an led_on message."""
    topic = build_topic(prefix=msg.ack, topic=LED_ON)
    kwargs = {}
    yield (topic, kwargs)


def led_off(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an led_off message."""
    topic = build_topic(prefix=msg.ack, topic=LED_OFF)
    kwargs = {}
    yield (topic, kwargs)


def manage_all_link_record(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an manage_all_link_record message."""
    topic = build_topic(prefix=msg.ack, topic=MANAGE_ALL_LINK_RECORD)
    kwargs = {
        "action": msg.action,
        "flags": msg.flags,
        "group": msg.group,
        "target": msg.target,
        "data1": msg.data1,
        "data2": msg.data2,
        "data3": msg.data3,
    }
    yield (topic, kwargs)


def set_nak_message_byte(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an set_nak_message_byte message."""
    topic = build_topic(prefix=msg.ack, topic=SET_NAK_MESSAGE_BYTE)
    kwargs = {"cmd2": msg.cmd2}
    yield (topic, kwargs)


def set_ack_message_two_bytes(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an set_ack_message_two_bytes message."""
    topic = build_topic(prefix=msg.ack, topic=SET_ACK_MESSAGE_TWO_BYTES)
    kwargs = {"cmd1": msg.cmd1, "cmd2": msg.cmd2}
    yield (topic, kwargs)


def rf_sleep(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an rf_sleep message."""
    topic = build_topic(prefix=msg.ack, topic=RF_SLEEP)
    kwargs = {}
    yield (topic, kwargs)


def get_im_configuration(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an get_im_configuration message."""
    topic = build_topic(prefix=msg.ack, topic=GET_IM_CONFIGURATION)
    if len(msg) == 3:
        kwargs = {}
    else:
        kwargs = {
            "disable_auto_linking": msg.flags.is_auto_link,
            "monitor_mode": msg.flags.is_monitor_mode,
            "auto_led": msg.flags.is_auto_led,
            "deadman": msg.flags.is_disable_deadman,
        }
    yield (topic, kwargs)


def read_eeprom(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from an read_eeprom message."""
    topic = build_topic(prefix=msg.ack, topic=READ_EEPROM)
    kwargs = {"mem_hi": msg.mem_hi, "mem_low": msg.mem_low}
    yield (topic, kwargs)


def write_eeprom(msg: Inbound) -> Tuple[str, Dict[str, Any]]:
    """Create a topic from a write eeprom message."""
    topic = build_topic(prefix=msg.ack, topic=WRITE_EEPROM)
    mem_addr = (msg.mem_hi << 8) + msg.mem_low + 7
    kwargs = {
        "mem_addr": mem_addr,
        "flags": msg.flags,
        "group": msg.group,
        "target": msg.target,
        "data1": msg.data1,
        "data2": msg.data2,
        "data3": msg.data3,
    }
    yield (topic, kwargs)


MSG_CONVERTER[0x50] = standard_received
MSG_CONVERTER[0x51] = extended_received
MSG_CONVERTER[0x52] = x10_received
MSG_CONVERTER[0x53] = all_linking_completed
MSG_CONVERTER[0x54] = button_event_report
MSG_CONVERTER[0x55] = user_reset_detected
MSG_CONVERTER[0x56] = all_link_cleanup_failure_report
MSG_CONVERTER[0x57] = all_link_record_response
MSG_CONVERTER[0x58] = all_link_cleanup_status_report
MSG_CONVERTER[0x59] = read_eeprom_response
MSG_CONVERTER[0x60] = get_im_info
MSG_CONVERTER[0x61] = send_all_link_command
MSG_CONVERTER[0x62] = send_standard_or_extended_message
MSG_CONVERTER[0x63] = x10_send
MSG_CONVERTER[0x64] = start_all_linking
MSG_CONVERTER[0x65] = cancel_all_linking
MSG_CONVERTER[0x66] = set_host_device_category
MSG_CONVERTER[0x67] = reset_im
MSG_CONVERTER[0x68] = set_ack_message_byte
MSG_CONVERTER[0x69] = get_first_all_link_record
MSG_CONVERTER[0x6A] = get_next_all_link_record
MSG_CONVERTER[0x6B] = set_im_configuration
MSG_CONVERTER[0x6C] = get_all_link_record_for_sender
MSG_CONVERTER[0x6D] = led_on
MSG_CONVERTER[0x6E] = led_off
MSG_CONVERTER[0x6F] = manage_all_link_record
MSG_CONVERTER[0x70] = set_nak_message_byte
MSG_CONVERTER[0x71] = set_ack_message_two_bytes
MSG_CONVERTER[0x72] = rf_sleep
MSG_CONVERTER[0x73] = get_im_configuration
MSG_CONVERTER[0x75] = read_eeprom
MSG_CONVERTER[0x76] = write_eeprom



"""KeypadLink radio button property."""
from itertools import chain

from . import OFF_MASK, ON_MASK, get_usable_value
from ..utils import bit_is_set
from .derived_property import DerivedProperty


def sort_groups(groups):
    """Sort the radio button groups from smallest to largest."""
    for group in groups:
        group.sort()
    groups.sort()
    return groups


def on_mask_name(button):
    """Return the name of the on mask for a button."""
    button_str = "" if button == 1 else f"_{button}"
    return f"{ON_MASK}{button_str}"


def off_mask_name(button):
    """Return the name of the off mask for a button."""
    button_str = "" if button == 1 else f"_{button}"
    return f"{OFF_MASK}{button_str}"


def calc_radio_button_groups(device, device_view=True):
    """Calculate the radio button groups based on the on masks.

    Set `device_view` to `True` to see the actual device setting.
    """
    rb_groups = []
    for button in device.groups:
        if button not in list(chain.from_iterable(rb_groups)):
            if device_view:
                on_mask = device.properties[on_mask_name(button)].value
            else:
                on_mask = get_usable_value(device.properties[on_mask_name(button)])
            if on_mask is None:
                on_mask = 0
            if on_mask != 0:
                rb_group = [button]
                for bit in list(range(0, button - 1)) + list(range(button, 8)):
                    if bit_is_set(on_mask, bit):
                        rb_group.append(bit + 1)
                if len(rb_group) > 1:
                    rb_groups.append(rb_group)
    rb_groups = sort_groups(rb_groups)
    return rb_groups


def set_radio_buttons(device, radio_button_groups):
    """Set the radio button groups in the on and off mask properties."""
    group_lists = []
    for group in radio_button_groups:
        group = list(group)
        if len(group) < 2:
            raise ValueError("Cannot have a radio button group with one button.")
        group = [int(button) for button in group]
        group.sort()
        group_lists.append(group)
    group_lists = sort_groups(group_lists)
    reset_radio_button_groups(device)
    for new_group in group_lists:
        for button in new_group:
            if button not in device.groups:
                raise ValueError(f"Invalid button value: {button}")
        group_dirty = True
        for group in calc_radio_button_groups(device):
            if group == new_group:
                group_dirty = False
        if group_dirty:
            device.set_radio_buttons(new_group)

    clear_buttons = []
    for button in device.groups:
        button_in_group = False
        for group in radio_button_groups:
            if button in group:
                button_in_group = True
        if not button_in_group:
            clear_buttons.append(button)
    if clear_buttons:
        device.clear_radio_buttons(clear_buttons)


def reset_radio_button_groups(device):
    """Clear any dirty on or off mask properties."""
    for button in device.groups:
        device.properties[on_mask_name(button)].new_value = None
        device.properties[off_mask_name(button)].new_value = None


class RadioButtonGroupsProperty(DerivedProperty):
    """KeypadLink radio button groups property."""

    def __init__(
        self,
        device,
        name,
    ):
        """Init the RadioButtonGroupProperty class."""
        super().__init__(device.address, name, list, False, False)
        self._device = device

    @property
    def value(self):
        """Return the value of the radio button groups from the underlying properties."""
        return calc_radio_button_groups(self._device, device_view=True)

    @property
    def new_value(self):
        """Return the modified value of the property."""
        new_value = calc_radio_button_groups(self._device, device_view=False)
        if new_value == self.value:
            return None
        return new_value

    @new_value.setter
    def new_value(self, value):
        """Modify the value of the property.

        Set value to a list of lists (i.e. [[2, 3], [4, 5]]).

        Set `new_value` to `None` to retain the current groups.

        Set `new_value` to an empty set (i.e. `[]`) to remove all groups.
        """

        if value is None:
            reset_radio_button_groups(self._device)
            return

        set_radio_buttons(self._device, value)

    @property
    def is_dirty(self):
        """Return if the underlying property values have changed."""
        on_mask_status = [
            self._device.properties[on_mask_name(group)].is_dirty
            for group in self._device.groups
        ]
        off_mask_status = [
            self._device.properties[off_mask_name(group)].is_dirty
            for group in self._device.groups
        ]
        return any(on_mask_status) or any(off_mask_status)

    @property
    def is_loaded(self):
        """Return if the underlying property values are loaded.."""
        on_mask_status = [
            self._device.properties[on_mask_name(group)].is_loaded
            for group in self._device.groups
        ]
        off_mask_status = [
            self._device.properties[off_mask_name(group)].is_loaded
            for group in self._device.groups
        ]
        return all(on_mask_status) and all(off_mask_status)



"""Peek one byte from device memory."""

from ...topics import PEEK
from .direct_command import DirectCommandHandlerBase


class PeekCommand(DirectCommandHandlerBase):
    """Peek one byte command."""

    def __init__(self, address):
        """Init the PeekCommand class."""
        super().__init__(topic=PEEK, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, lsb: int):
        """Send the PEEK command async."""
        return await super().async_send(lsb=lsb)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(value=cmd2)



"""Assign a device to an all link group."""
from ...address import Address
from ...topics import ASSIGN_TO_ALL_LINK_GROUP
from .broadcast_command import BroadcastCommandHandlerBase


class AssignToAllLinkGroupCommand(BroadcastCommandHandlerBase):
    """Assign to All-Link Group command handler."""

    def __init__(self, address: Address):
        """Init the AssignToAllLinkGroupCommand class."""
        super().__init__(topic=ASSIGN_TO_ALL_LINK_GROUP, address=address)

    def _handle_message_received(self, cmd1, cmd2, target, user_data, hops_left):
        """Receive the device ID information."""
        cat = target.high
        subcat = target.middle
        firmware = target.low
        self._call_subscribers(
            address=self._address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            group=cmd2,
            link_mode=None,
        )



"""Insteon Powerline Modem (PLM)."""

from ..aldb.modem_aldb import ModemALDB
from .modem_base import ModemBase


class MockModem(ModemBase):
    """Insteon PLM class."""

    def __init__(
        self,
        address="000000",
        cat=0x03,
        subcat=0x00,
        firmware=0x00,
        description="",
        model="",
    ):
        """Init the Modem class."""
        super().__init__(address, cat, subcat, firmware, description, model)
        self._aldb = ModemALDB(self._address, mem_addr=0x1FFF)



"""Peek one byte from device memory."""

from ...topics import POKE
from .direct_command import DirectCommandHandlerBase


class PokeCommand(DirectCommandHandlerBase):
    """Poke one byte command."""

    def __init__(self, address):
        """Init the PokeCommand class."""
        super().__init__(topic=POKE, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, value: int):
        """Send the POKE command async."""
        return await super().async_send(value=value)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(value=cmd2)



"""Create outbound messages."""
import logging

from ... import pub
from ...address import Address
from ...constants import (
    AllLinkMode,
    DeviceCategory,
    ManageAllLinkRecordAction,
    MessageId,
    X10CommandType,
)
from ...data_types.all_link_record_flags import AllLinkRecordFlags
from ...data_types.im_config_flags import IMConfigurationFlags
from ...data_types.message_flags import MessageFlags
from ...data_types.user_data import UserData
from ...topics import (
    CANCEL_ALL_LINKING,
    GET_ALL_LINK_RECORD_FOR_SENDER,
    GET_FIRST_ALL_LINK_RECORD,
    GET_IM_CONFIGURATION,
    GET_IM_INFO,
    GET_NEXT_ALL_LINK_RECORD,
    LED_OFF,
    LED_ON,
    MANAGE_ALL_LINK_RECORD,
    READ_EEPROM,
    RESET_IM,
    RF_SLEEP,
    SEND_ALL_LINK_COMMAND,
    SEND_EXTENDED,
    SEND_STANDARD,
    SET_ACK_MESSAGE_BYTE,
    SET_ACK_MESSAGE_TWO_BYTES,
    SET_HOST_DEVICE_CATEGORY,
    SET_IM_CONFIGURATION,
    SET_NAK_MESSAGE_BYTE,
    START_ALL_LINKING,
    WRITE_EEPROM,
    X10_SEND,
)
from ...utils import subscribe_topic
from ..topic_converters import topic_to_message_handler, topic_to_message_type
from . import MessageBase
from .message_definition import MessageDefinition
from .message_definitions import FLD_EXT_SEND, FLD_STD_SEND, OUTBOUND_MSG_DEF

MESSAGE_REGISTER = {}

_LOGGER = logging.getLogger(__name__)


def register_outbound_handlers():
    """Register outbound handlers."""
    for topic, func in MESSAGE_REGISTER.items():
        subscribe_topic(func, topic)


class OutboundWriteManager:
    """Ourbound write manager."""

    def __init__(self):
        """Init the OutboundWriteManager class."""
        self._protocol_write = None

    @property
    def protocol_write(self):
        """Return the write method of the protocol."""
        return self._protocol_write

    @protocol_write.setter
    def protocol_write(self, value):
        """Set the write method of the protocol."""
        self._protocol_write = value

    def write(self, msg, priority):
        """Write to the protocol."""
        if self._protocol_write is None:
            raise AttributeError
        self._protocol_write(msg=msg, priority=priority)


outbound_write_manager = OutboundWriteManager()


class Outbound(MessageBase):
    """Outbound message class."""

    def __eq__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns True.
        """
        return True

    def __gt__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns False.
        """
        return False

    def __lt__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns False.
        """
        return False

    def __hash__(self):
        """Needed to allow a prioritized queue work.

        Return hash of str(self).
        """
        return hash(str(self))


def _create_outbound_message(topic, priority=5, **kwargs) -> Outbound:
    """Create an Outbound message."""
    topic = topic.name.split(".")[1]
    msg_id = getattr(MessageId, topic.upper())
    msg_def = OUTBOUND_MSG_DEF[msg_id]
    msg = Outbound(msg_def, **kwargs)
    outbound_write_manager.write(msg=msg, priority=priority)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=GET_IM_INFO)
def get_im_info(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_IM_INFO outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SEND_ALL_LINK_COMMAND)
def send_all_link_command(
    group: int, cmd1: int, cmd2: int, topic=pub.AUTO_TOPIC
) -> Outbound:
    """Create a SEND_ALL_LINK_COMMAND outbound message."""
    _create_outbound_message(group=group, cmd1=cmd1, cmd2=cmd2, topic=topic, priority=3)


def _create_flags(topic, extended):
    msg_type = topic_to_message_type(topic)
    return MessageFlags.create(msg_type, extended=extended)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SEND_STANDARD)
def send_standard(
    address: Address,
    cmd1: int,
    cmd2: int,
    flags: MessageFlags = None,
    priority=5,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SEND_STANDARD outbound message."""
    flags = flags if flags is not None else _create_flags(topic, False)
    kwargs = {"address": address, "flags": flags, "cmd1": cmd1, "cmd2": cmd2}
    msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_STD_SEND)
    msg = Outbound(msg_def, **kwargs)
    outbound_write_manager.write(msg=msg, priority=priority)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SEND_EXTENDED)
def send_extended(
    address: Address,
    cmd1: int,
    cmd2: int,
    user_data: UserData,
    flags: MessageFlags = None,
    priority=5,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SEND_EXTENDED outbound message."""
    flags = flags if flags is not None else _create_flags(topic, True)
    kwargs = {
        "address": address,
        "flags": flags,
        "cmd1": cmd1,
        "cmd2": cmd2,
        "user_data": user_data,
    }
    msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_EXT_SEND)
    msg = Outbound(msg_def, **kwargs)
    outbound_write_manager.write(msg=msg, priority=priority)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=X10_SEND)
def x10_send(raw_x10: int, x10_flag: X10CommandType, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a X10_SEND outbound message."""
    _create_outbound_message(raw_x10=raw_x10, x10_flag=x10_flag, topic=topic)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=START_ALL_LINKING)
def start_all_linking(
    link_mode: AllLinkMode, group: int, topic=pub.AUTO_TOPIC
) -> Outbound:
    """Create a START_ALL_LINKING outbound message."""
    _create_outbound_message(link_mode=link_mode, group=group, topic=topic, priority=7)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=CANCEL_ALL_LINKING)
def cancel_all_linking(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a CANCEL_ALL_LINKING outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(
    register_list=MESSAGE_REGISTER, topic=SET_HOST_DEVICE_CATEGORY
)
def set_host_device_category(
    cat: DeviceCategory, subcat: int, firmware: int, topic=pub.AUTO_TOPIC
) -> Outbound:
    """Create a SET_HOST_DEVICE_CATEGORY outbound message."""
    _create_outbound_message(
        cat=cat, subcat=subcat, firmware=firmware, topic=topic, priority=7
    )


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=RESET_IM)
def reset_im(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a RESET_IM outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SET_ACK_MESSAGE_BYTE)
def set_ack_message_byte(cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_ACK_MESSAGE_BYTE outbound message."""
    _create_outbound_message(cmd2=cmd2, topic=topic, priority=7)


@topic_to_message_handler(
    register_list=MESSAGE_REGISTER, topic=GET_FIRST_ALL_LINK_RECORD
)
def get_first_all_link_record(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_FIRST_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(
    register_list=MESSAGE_REGISTER, topic=GET_NEXT_ALL_LINK_RECORD
)
def get_next_all_link_record(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_NEXT_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SET_IM_CONFIGURATION)
def set_im_configuration(
    disable_auto_linking: bool,
    monitor_mode: bool,
    auto_led: bool,
    deadman: bool,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SET_IM_CONFIGURATION outbound message."""
    flag_byte = 0x00
    flag_byte = flag_byte | (1 if disable_auto_linking else 0) << 7
    flag_byte = flag_byte | (1 if monitor_mode else 0) << 6
    flag_byte = flag_byte | (1 if auto_led else 0) << 5
    flag_byte = flag_byte | (1 if deadman else 0) << 4
    flags = IMConfigurationFlags(flag_byte)
    _create_outbound_message(flags=flags, topic=topic, priority=2)


@topic_to_message_handler(
    register_list=MESSAGE_REGISTER, topic=GET_ALL_LINK_RECORD_FOR_SENDER
)
def get_all_link_record_for_sender(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_ALL_LINK_RECORD_FOR_SENDER outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=LED_ON)
def led_on(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a LED_ON outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=LED_OFF)
def led_off(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a LED_OFF outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=MANAGE_ALL_LINK_RECORD)
def manage_all_link_record(
    action: ManageAllLinkRecordAction,
    flags: AllLinkRecordFlags,
    group: int,
    target: Address,
    data1: int,
    data2: int,
    data3: int,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a MANAGE_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(
        action=action,
        flags=flags,
        group=group,
        target=target,
        data1=data1,
        data2=data2,
        data3=data3,
        topic=topic,
        priority=10,
    )


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=SET_NAK_MESSAGE_BYTE)
def set_nak_message_byte(cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_NAK_MESSAGE_BYTE outbound message."""
    _create_outbound_message(cmd2=cmd2, topic=topic, priority=10)


@topic_to_message_handler(
    register_list=MESSAGE_REGISTER, topic=SET_ACK_MESSAGE_TWO_BYTES
)
def set_ack_message_two_bytes(cmd1: int, cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_ACK_MESSAGE_TWO_BYTES outbound message."""
    _create_outbound_message(cmd1=cmd1, cmd2=cmd2, topic=topic, priority=10)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=RF_SLEEP)
def rf_sleep(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a RF_SLEEP outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=GET_IM_CONFIGURATION)
def get_im_configuration(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_IM_CONFIGURATION outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=READ_EEPROM)
def read_eeprom(mem_hi: int, mem_low: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a READ_EEPROM outbound message."""
    _create_outbound_message(mem_hi=mem_hi, mem_low=mem_low, topic=topic, priority=10)


@topic_to_message_handler(register_list=MESSAGE_REGISTER, topic=WRITE_EEPROM)
def write_eeprom(
    mem_hi: int,
    mem_low: int,
    flags: AllLinkRecordFlags,
    group: int,
    target: Address,
    data1: int,
    data2: int,
    data3: int,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a WRITE_EEPROM outbound message."""
    _create_outbound_message(
        mem_hi=mem_hi,
        mem_low=mem_low,
        flags=flags,
        group=group,
        target=target,
        data1=data1,
        data2=data2,
        data3=data3,
        topic=topic,
        priority=10,
    )
